=== added file 'DBUS-API'
--- DBUS-API	1970-01-01 00:00:00 +0000
+++ DBUS-API	2010-09-26 17:36:30 +0000
@@ -0,0 +1,151 @@
+		   -*- mode: org; coding: utf-8 -*-
+
+		    Mandos Server D-Bus Interface
+
+This file documents the D-Bus interface to the Mandos server.
+
+* Bus: System bus
+  Bus name: "se.bsnet.fukt.Mandos"
+
+
+* Object Paths:
+  
+  | Path                  | Object            |
+  |-----------------------+-------------------|
+  | "/"                   | The Mandos Server |
+  | "/clients/CLIENTNAME" | Mandos Client     |
+
+  
+* Mandos Server Interface:
+  Interface name: "se.bsnet.fukt.Mandos"
+  
+** Methods:
+*** GetAllClients() → (ao: Clients)
+    Returns an array of all client D-Bus object paths
+   
+*** GetAllClientsWithProperties() → (a{oa{sv}}: ClientProperties)
+    Returns an array of all clients and all their properties
+   
+*** RemoveClient(o: ObjectPath) → nothing
+    Removes a client
+   
+** Signals:
+*** ClientAdded(o: ObjectPath)
+    A new client was added.
+   
+*** ClientNotFound(s: Fingerprint, s: Address)
+    A client connected from Address using Fingerprint, but was
+    rejected because it was not found in the server.  The fingerprint
+    is represented as a string of hexadecimal digits.  The address is
+    an IPv4 or IPv6 address in its normal string format.
+   
+*** ClientRemoved(o: ObjectPath, s: Name)
+    A client named Name on ObjectPath was removed.
+
+
+* Mandos Client Interface:
+  Interface name: "se.bsnet.fukt.Mandos.Client"
+  
+** Methods
+*** Approve(b: Approve) → nothing
+    Approve or deny a connected client waiting for approval.  If
+    denied, a client will not be sent its secret.
+    
+*** CheckedOK() → nothing
+    Assert that this client has been checked and found to be alive.
+    This will restart the timeout before disabling this client.  See
+    also the "LastCheckedOK" property.
+    
+*** Disable() → nothing
+    Disable this client.  See also the "Enabled" property.
+    
+*** Enable() → nothing
+    Enable this client.  See also the "Enabled" property.
+    
+*** StartChecker() → nothing
+    Start a new checker for this client, if none is currently
+    running.  See also the "CheckerRunning" property.
+    
+*** StopChecker() → nothing
+    Abort a running checker process for this client, if any.  See also
+    the "CheckerRunning" property.
+
+** Properties
+   
+   Note: Many of these properties directly correspond to a setting in
+   "clients.conf", in which case they are fully documented in
+   mandos-clients.conf(5).
+   
+   | Name                    | Type | Access     | clients.conf        |
+   |-------------------------+------+------------+---------------------|
+   | ApprovedByDefault       | b    | Read/Write | approved_by_default |
+   | ApprovalDelay (a)       | t    | Read/Write | approval_delay      |
+   | ApprovalDuration (a)    | t    | Read/Write | approval_duration   |
+   | ApprovalPending (b)     | b    | Read       | N/A                 |
+   | Checker                 | s    | Read/Write | checker             |
+   | CheckerRunning (c)      | b    | Read/Write | N/A                 |
+   | Created (d)             | s    | Read       | N/A                 |
+   | Enabled (e)             | b    | Read/Write | N/A                 |
+   | Fingerprint             | s    | Read       | fingerprint         |
+   | Host                    | s    | Read/Write | host                |
+   | Interval (a)            | t    | Read/Write | interval            |
+   | LastApprovalRequest (f) | s    | Read       | N/A                 |
+   | LastCheckedOK (g)       | s    | Read/Write | N/A                 |
+   | LastEnabled (h)         | s    | Read       | N/A                 |
+   | Name                    | s    | Read       | (Section name)      |
+   | ObjectPath              | o    | Read       | N/A                 |
+   | Secret (i)              | ay   | Write      | secret (or secfile) |
+   | Timeout (a)             | t    | Read/Write | timeout             |
+   
+   a) Represented as milliseconds.
+   
+   b) An approval is currently pending.
+   
+   c) Setting this property is equivalent to calling StartChecker() or
+      StopChecker().
+   
+   d) The creation time of this client object, as a RFC 3339 string.
+   
+   e) Setting this property is equivalent to calling Enable() or
+      Disable().
+   
+   f) The time of the last approval request, as a RFC 3339 string, or
+      an empty string if this has not happened.
+   
+   g) The last time a checker was successful, as a RFC 3339 string, or
+      an empty string if this has not happened.  Setting this property
+      is equivalent to calling CheckedOK(), i.e. the current time is
+      set, regardless of the string sent.  Please always use an empty
+      string when setting this property, to allow for possible future
+      expansion.
+   
+   h) The last time this client was enabled, as a RFC 3339 string, or
+      an empty string if this has not happened.
+   
+   i) A raw byte array, not hexadecimal digits.
+
+** Signals
+*** CheckerCompleted(n: Exitcode, x: Waitstatus, s: Command)
+    A checker (Command) has completed.  Exitcode is either the exit
+    code or -1 for abnormal exit.  In any case, the full Waitstatus
+    (as from wait(2)) is also available.
+    
+*** CheckerStarted(s: Command)
+    A checker command (Command) has just been started.
+    
+*** GotSecret()
+    This client has been sent its secret.
+    
+*** NeedApproval(t: Timeout, b: ApprovedByDefault)
+    This client will be approved or denied in exactly Timeout
+    milliseconds, depending on ApprovedByDefault.  Approve() can now
+    usefully be called on this client object.
+
+*** PropertyChanged(s: Property, v: Value)
+    The Property on this client has changed to Value.
+
+*** Rejected(s: Reason)
+    This client was not given its secret for a specified Reason.
+
+
+#+STARTUP: showall

=== modified file 'Makefile'
--- Makefile	2010-09-10 17:06:26 +0000
+++ Makefile	2010-09-23 20:14:08 +0000
@@ -83,14 +83,16 @@
 	--expression='s/\(<a class="citerefentry" href="\)\("><span class="citerefentry"><span class="refentrytitle">\)\([^<]*\)\(<\/span>(\)\([^)]*\)\()<\/span><\/a>\)/\1\3.\5\2\3\4\5\6/g'
 
 PLUGINS=plugins.d/password-prompt plugins.d/mandos-client \
-	plugins.d/usplash plugins.d/splashy plugins.d/askpass-fifo
+	plugins.d/usplash plugins.d/splashy plugins.d/askpass-fifo \
+	plugins.d/plymouth
 CPROGS=plugin-runner $(PLUGINS)
-PROGS=mandos mandos-keygen mandos-ctl $(CPROGS)
-DOCS=mandos.8 plugin-runner.8mandos mandos-keygen.8 \
+PROGS=mandos mandos-keygen mandos-ctl mandos-monitor $(CPROGS)
+DOCS=mandos.8 mandos-keygen.8 mandos-monitor.8 mandos-ctl.8 \
+	mandos.conf.5 mandos-clients.conf.5 plugin-runner.8mandos \
 	plugins.d/mandos-client.8mandos \
-	plugins.d/password-prompt.8mandos mandos.conf.5 \
-	plugins.d/usplash.8mandos plugins.d/splashy.8mandos \
-	plugins.d/askpass-fifo.8mandos mandos-clients.conf.5
+	plugins.d/password-prompt.8mandos plugins.d/usplash.8mandos \
+	plugins.d/splashy.8mandos plugins.d/askpass-fifo.8mandos \
+	plugins.d/plymouth.8mandos
 
 htmldocs=$(addsuffix .xhtml,$(DOCS))
 
@@ -131,6 +133,20 @@
 		 legalnotice.xml
 	$(DOCBOOKTOHTML)
 
+mandos-monitor.8: mandos-monitor.xml common.ent overview.xml \
+		legalnotice.xml
+	$(DOCBOOKTOMAN)
+mandos-monitor.8.xhtml: mandos-monitor.xml common.ent overview.xml \
+		 legalnotice.xml
+	$(DOCBOOKTOHTML)
+
+mandos-ctl.8: mandos-ctl.xml common.ent overview.xml \
+		legalnotice.xml
+	$(DOCBOOKTOMAN)
+mandos-ctl.8.xhtml: mandos-ctl.xml common.ent overview.xml \
+		 legalnotice.xml
+	$(DOCBOOKTOHTML)
+
 mandos.conf.5: mandos.conf.xml common.ent mandos-options.xml \
 		legalnotice.xml
 	$(DOCBOOKTOMAN)
@@ -177,6 +193,11 @@
 		--expression='s/^\(version = "\)[^"]*"$$/\1$(version)"/' \
 		$@)
 
+mandos-monitor: Makefile
+	$(strip $(SED) --in-place \
+		--expression='s/^\(version = "\)[^"]*"$$/\1$(version)"/' \
+		$@)
+
 mandos.lsm: Makefile
 	$(strip $(SED) --in-place \
 		--expression='s/^\(Version:\).*/\1\t$(version)/' \
@@ -233,11 +254,10 @@
 # Run the server with a local config
 run-server: confdir/mandos.conf confdir/clients.conf
 	@echo "#################################################################"
-	@echo "# NOTE: Please IGNORE errors about \"No permission to bind to    #"
-	@echo "# interface\" or \"Could not open file u'/var/run/mandos.pid'\" -  #"
-	@echo "# they are harmless and are caused by the server not running as #"
-	@echo "# root.  Do NOT run \"make run-server\" server as root if you did #"
-	@echo "# not also unpack and compile it as root.                       #"
+	@echo "# NOTE: Please IGNORE the error about \"Could not open file      #"
+	@echo "# u'/var/run/mandos.pid'\" -  it is harmless and is caused by    #"
+	@echo "# the server not running as root.  Do NOT run \"make run-server\" #"
+	@echo "# server as root if you didn't also unpack and compile it thus. #"
 	@echo "#################################################################"
 	./mandos --debug --no-dbus --configdir=confdir $(SERVERARGS)
 
@@ -261,6 +281,10 @@
 install-server: doc
 	install --directory $(CONFDIR)
 	install --mode=u=rwx,go=rx mandos $(PREFIX)/sbin/mandos
+	install --mode=u=rwx,go=rx --target-directory=$(PREFIX)/sbin \
+		mandos-ctl
+	install --mode=u=rwx,go=rx --target-directory=$(PREFIX)/sbin \
+		mandos-monitor
 	install --mode=u=rw,go=r --target-directory=$(CONFDIR) \
 		mandos.conf
 	install --mode=u=rw --target-directory=$(CONFDIR) \
@@ -276,6 +300,10 @@
 	fi
 	gzip --best --to-stdout mandos.8 \
 		> $(MANDIR)/man8/mandos.8.gz
+	gzip --best --to-stdout mandos-monitor.8 \
+		> $(MANDIR)/man8/mandos-monitor.8.gz
+	gzip --best --to-stdout mandos-ctl.8 \
+		> $(MANDIR)/man8/mandos-ctl.8.gz
 	gzip --best --to-stdout mandos.conf.5 \
 		> $(MANDIR)/man5/mandos.conf.5.gz
 	gzip --best --to-stdout mandos-clients.conf.5 \
@@ -308,6 +336,9 @@
 	install --mode=u=rwxs,go=rx \
 		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
 		plugins.d/askpass-fifo
+	install --mode=u=rwxs,go=rx \
+		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		plugins.d/plymouth
 	install initramfs-tools-hook \
 		$(INITRAMFSTOOLS)/hooks/mandos
 	install --mode=u=rw,go=r initramfs-tools-hook-conf \
@@ -319,16 +350,18 @@
 		> $(MANDIR)/man8/mandos-keygen.8.gz
 	gzip --best --to-stdout plugin-runner.8mandos \
 		> $(MANDIR)/man8/plugin-runner.8mandos.gz
+	gzip --best --to-stdout plugins.d/mandos-client.8mandos \
+		> $(MANDIR)/man8/mandos-client.8mandos.gz
 	gzip --best --to-stdout plugins.d/password-prompt.8mandos \
 		> $(MANDIR)/man8/password-prompt.8mandos.gz
-	gzip --best --to-stdout plugins.d/mandos-client.8mandos \
-		> $(MANDIR)/man8/mandos-client.8mandos.gz
 	gzip --best --to-stdout plugins.d/usplash.8mandos \
 		> $(MANDIR)/man8/usplash.8mandos.gz
 	gzip --best --to-stdout plugins.d/splashy.8mandos \
 		> $(MANDIR)/man8/splashy.8mandos.gz
 	gzip --best --to-stdout plugins.d/askpass-fifo.8mandos \
 		> $(MANDIR)/man8/askpass-fifo.8mandos.gz
+	gzip --best --to-stdout plugins.d/plymouth.8mandos \
+		> $(MANDIR)/man8/plymouth.8mandos.gz
 
 install-client: install-client-nokey
 # Post-installation stuff
@@ -340,7 +373,11 @@
 
 uninstall-server:
 	-rm --force $(PREFIX)/sbin/mandos \
+		$(PREFIX)/sbin/mandos-ctl \
+		$(PREFIX)/sbin/mandos-monitor \
 		$(MANDIR)/man8/mandos.8.gz \
+		$(MANDIR)/man8/mandos-monitor.8.gz \
+		$(MANDIR)/man8/mandos-ctl.8.gz \
 		$(MANDIR)/man5/mandos.conf.5.gz \
 		$(MANDIR)/man5/mandos-clients.conf.5.gz
 	update-rc.d -f mandos remove
@@ -358,16 +395,18 @@
 		$(PREFIX)/lib/mandos/plugins.d/usplash \
 		$(PREFIX)/lib/mandos/plugins.d/splashy \
 		$(PREFIX)/lib/mandos/plugins.d/askpass-fifo \
+		$(PREFIX)/lib/mandos/plugins.d/plymouth \
 		$(INITRAMFSTOOLS)/hooks/mandos \
 		$(INITRAMFSTOOLS)/conf-hooks.d/mandos \
 		$(INITRAMFSTOOLS)/scripts/init-premount/mandos \
+		$(MANDIR)/man8/mandos-keygen.8.gz \
 		$(MANDIR)/man8/plugin-runner.8mandos.gz \
-		$(MANDIR)/man8/mandos-keygen.8.gz \
+		$(MANDIR)/man8/mandos-client.8mandos.gz
 		$(MANDIR)/man8/password-prompt.8mandos.gz \
 		$(MANDIR)/man8/usplash.8mandos.gz \
 		$(MANDIR)/man8/splashy.8mandos.gz \
 		$(MANDIR)/man8/askpass-fifo.8mandos.gz \
-		$(MANDIR)/man8/mandos-client.8mandos.gz
+		$(MANDIR)/man8/plymouth.8mandos.gz \
 	-rmdir $(PREFIX)/lib/mandos/plugins.d $(CONFDIR)/plugins.d \
 		 $(PREFIX)/lib/mandos $(CONFDIR) $(KEYDIR)
 	update-initramfs -k all -u

=== modified file 'TODO'
--- TODO	2010-09-09 18:16:14 +0000
+++ TODO	2010-09-26 17:36:30 +0000
@@ -2,10 +2,12 @@
 
 * Use _attribute_((nonnull)) wherever possible.
 
+* Release critical
+** Update copyright year to add 2010
+
 * mandos-client
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO use error() instead of perror()
 ** TODO [#B] Retry a server which has a non-definite reply:
 *** A closed connection during the TLS handshake
 *** A TCP timeout
@@ -15,22 +17,19 @@
 * splashy
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 
 * usplash
 ** TODO [#A] Make it work again
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
+** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
 
 * askpass-fifo
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] Drop privileges after opening FIFO.
 
 * password-prompt
 ** TODO [#B] Prefix all debug output with "Mandos plugin " + program_invocation_short_name
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] lock stdin (with flock()?)
 
 * TODO [#B] passdev
@@ -39,7 +38,6 @@
 ** TODO [#B] use scandir(3) instead of readdir(3)
 ** TODO [#C] use same file name rules as run-parts(8)
 ** kernel command line option for debug info
-** TODO [#B] use error() instead of perror()
 ** TODO [#B] Use openat()
 
 * mandos (server)
@@ -82,14 +80,16 @@
 ** TODO [#C] python-parsedatetime
 ** TODO [#C] systemd/launchd
    http://0pointer.de/blog/projects/systemd.html
+** TODO Separate logging logic to own object
+** TODO make clients to a dict!
+** TODO [#A] Limit approval_delay to max gnutls/tls timeout value
+** TODO [#B] break the wait on approval_delay if connection dies
+** TODO Generate Client.runtime_expansions from client options + extra
+** TODO Allow %%(checker)s as a runtime expansion
 
 * mandos.xml
-** [[file:mandos.xml::XXX][Document D-Bus interface]]
-   Remove mention of lack of such interface in BUGS section
 ** Add mandos contact info in manual pages
 
-* TODO [#A] Provide and install /etc/dbus-1/system.d/mandos.conf
-
 * mandos-ctl
 *** Handle "no D-Bus server" and/or "no Mandos server found" better
 *** [#B] --dump option
@@ -100,9 +100,13 @@
   arguments.
 
 * mandos-monitor
+** TODO help should be toggable
 ** Urwid client data displayer
    Better view of client data in the listing
 *** Properties popup
+** Nicer crashes. Stack traces Messes up shell.
+*** Print a nice "We are sorry" message, save stack trace to log.
+** Show timeout countdown for approval
 
 * mandos-keygen
 ** TODO Loop until passwords match when run interactively
@@ -125,5 +129,9 @@
 ** TODO [#C] /etc/bash_completion.d/mandos
    From XML sources directly?
 
+* Side Stuff
+** TODO Locate which packet move the other bin/sh when busy box is deactivated
+** TODO contact owner of packet, and ask them to have that shell static in position regardless of busybox
+
 
 #+STARTUP: showall

=== modified file 'clients.conf'
--- clients.conf	2009-12-25 23:13:47 +0000
+++ clients.conf	2010-09-12 03:00:40 +0000
@@ -16,6 +16,15 @@
 # What command to run as "the checker".
 ;checker = fping -q -- %%(host)s
 
+# Whether to approve a client by default after the approval delay.
+;approved_by_default = True
+
+# How long to wait for approval.
+;approval_delay = 0s
+
+# How long one approval will last.
+;approval_duration = 1s
+
 
 ;####
 ;# Example client
@@ -62,4 +71,9 @@
 ;
 ;# Parameters from the [DEFAULT] section can be overridden per client.
 ;interval = 5m
+;
+;# This client requires manual approval before it receives its secret.
+;approved_by_default = False
+;# Require approval within 30 seconds.
+;approval_delay = 30s
 ;####

=== modified file 'debian/control'
--- debian/control	2010-09-09 18:16:14 +0000
+++ debian/control	2010-09-09 22:26:35 +0000
@@ -15,7 +15,8 @@
 Package: mandos
 Architecture: all
 Depends: ${misc:Depends}, python (>=2.5), python-gnutls, python-dbus,
-	python-avahi, python-gobject, avahi-daemon, adduser
+	python-avahi, python-gobject, avahi-daemon, adduser,
+	python-urwid
 Recommends: fping
 Description: a server giving encrypted passwords to Mandos clients
  This is the server part of the Mandos system, which allows

=== modified file 'debian/mandos.dirs'
--- debian/mandos.dirs	2008-09-17 00:34:09 +0000
+++ debian/mandos.dirs	2010-09-15 17:33:14 +0000
@@ -2,4 +2,5 @@
 usr/share/man/man8
 etc/init.d
 etc/default
+etc/dbus-1/system.d
 usr/sbin

=== modified file 'debian/mandos.docs'
--- debian/mandos.docs	2008-10-18 11:17:22 +0000
+++ debian/mandos.docs	2010-09-12 03:00:40 +0000
@@ -1,3 +1,4 @@
 NEWS
 README
 TODO
+DBUS-API

=== modified file 'debian/watch'
--- debian/watch	2009-01-15 02:52:02 +0000
+++ debian/watch	2010-09-15 17:17:46 +0000
@@ -1,2 +1,2 @@
 version=3
-ftp://ftp.fukt.bsnet.se/pub/mandos/mandos[-_]([^\s]+?)(?:\.orig)?\.tar\.(?:gz|bz2|7z)
+ftp://ftp.fukt.bsnet.se/pub/mandos/mandos[-_]([^\s]+?)(?:\.orig)?\.tar\.(?:gz|bz2|7z|xz)

=== modified file 'mandos'
--- mandos	2010-09-09 18:16:14 +0000
+++ mandos	2010-09-26 17:36:30 +0000
@@ -60,7 +60,7 @@
 import fcntl
 import functools
 import cPickle as pickle
-import select
+import multiprocessing
 
 import dbus
 import dbus.service
@@ -83,6 +83,7 @@
 
 version = "1.0.14"
 
+#logger = logging.getLogger(u'mandos')
 logger = logging.Logger(u'mandos')
 syslogger = (logging.handlers.SysLogHandler
              (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
@@ -156,15 +157,20 @@
                             u" after %i retries, exiting.",
                             self.rename_count)
             raise AvahiServiceError(u"Too many renames")
-        self.name = self.server.GetAlternativeServiceName(self.name)
+        self.name = unicode(self.server.GetAlternativeServiceName(self.name))
         logger.info(u"Changing Zeroconf service name to %r ...",
-                    unicode(self.name))
+                    self.name)
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
                                 u' %%(levelname)s: %%(message)s'
                                 % self.name))
         self.remove()
-        self.add()
+        try:
+            self.add()
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            self.cleanup()
+            os._exit(1)
         self.rename_count += 1
     def remove(self):
         """Derived from the Avahi example code"""
@@ -234,33 +240,43 @@
     """A representation of a client host served by this server.
     
     Attributes:
-    name:       string; from the config file, used in log messages and
-                        D-Bus identifiers
+    _approved:   bool(); 'None' if not yet approved/disapproved
+    approval_delay: datetime.timedelta(); Time to wait for approval
+    approval_duration: datetime.timedelta(); Duration of one approval
+    checker:    subprocess.Popen(); a running checker process used
+                                    to see if the client lives.
+                                    'None' if no process is running.
+    checker_callback_tag: a gobject event source tag, or None
+    checker_command: string; External command which is run to check
+                     if client lives.  %() expansions are done at
+                     runtime with vars(self) as dict, so that for
+                     instance %(name)s can be used in the command.
+    checker_initiator_tag: a gobject event source tag, or None
+    created:    datetime.datetime(); (UTC) object creation
+    current_checker_command: string; current running checker_command
+    disable_hook:  If set, called by disable() as disable_hook(self)
+    disable_initiator_tag: a gobject event source tag, or None
+    enabled:    bool()
     fingerprint: string (40 or 32 hexadecimal digits); used to
                  uniquely identify the client
-    secret:     bytestring; sent verbatim (over TLS) to client
     host:       string; available for use by the checker command
-    created:    datetime.datetime(); (UTC) object creation
+    interval:   datetime.timedelta(); How often to start a new checker
+    last_approval_request: datetime.datetime(); (UTC) or None
+    last_checked_ok: datetime.datetime(); (UTC) or None
     last_enabled: datetime.datetime(); (UTC)
-    enabled:    bool()
-    last_checked_ok: datetime.datetime(); (UTC) or None
+    name:       string; from the config file, used in log messages and
+                        D-Bus identifiers
+    secret:     bytestring; sent verbatim (over TLS) to client
     timeout:    datetime.timedelta(); How long from last_checked_ok
                                       until this client is disabled
-    interval:   datetime.timedelta(); How often to start a new checker
-    disable_hook:  If set, called by disable() as disable_hook(self)
-    checker:    subprocess.Popen(); a running checker process used
-                                    to see if the client lives.
-                                    'None' if no process is running.
-    checker_initiator_tag: a gobject event source tag, or None
-    disable_initiator_tag: - '' -
-    checker_callback_tag:  - '' -
-    checker_command: string; External command which is run to check if
-                     client lives.  %() expansions are done at
-                     runtime with vars(self) as dict, so that for
-                     instance %(name)s can be used in the command.
-    current_checker_command: string; current running checker_command
+    runtime_expansions: Allowed attributes for runtime expansion.
     """
     
+    runtime_expansions = (u"approval_delay", u"approval_duration",
+                          u"created", u"enabled", u"fingerprint",
+                          u"host", u"interval", u"last_checked_ok",
+                          u"last_enabled", u"name", u"timeout")
+    
     @staticmethod
     def _timedelta_to_milliseconds(td):
         "Convert a datetime.timedelta() to milliseconds"
@@ -275,6 +291,9 @@
     def interval_milliseconds(self):
         "Return the 'interval' attribute in milliseconds"
         return self._timedelta_to_milliseconds(self.interval)
+
+    def approval_delay_milliseconds(self):
+        return self._timedelta_to_milliseconds(self.approval_delay)
     
     def __init__(self, name = None, disable_hook=None, config=None):
         """Note: the 'checker' key in 'config' sets the
@@ -303,6 +322,7 @@
         self.host = config.get(u"host", u"")
         self.created = datetime.datetime.utcnow()
         self.enabled = False
+        self.last_approval_request = None
         self.last_enabled = None
         self.last_checked_ok = None
         self.timeout = string_to_delta(config[u"timeout"])
@@ -315,12 +335,27 @@
         self.checker_command = config[u"checker"]
         self.current_checker_command = None
         self.last_connect = None
+        self._approved = None
+        self.approved_by_default = config.get(u"approved_by_default",
+                                              True)
+        self.approvals_pending = 0
+        self.approval_delay = string_to_delta(
+            config[u"approval_delay"])
+        self.approval_duration = string_to_delta(
+            config[u"approval_duration"])
+        self.changedstate = multiprocessing_manager.Condition(multiprocessing_manager.Lock())
     
+    def send_changedstate(self):
+        self.changedstate.acquire()
+        self.changedstate.notify_all()
+        self.changedstate.release()
+        
     def enable(self):
         """Start this client's checker and timeout hooks"""
         if getattr(self, u"enabled", False):
             # Already enabled
             return
+        self.send_changedstate()
         self.last_enabled = datetime.datetime.utcnow()
         # Schedule a new checker to be started an 'interval' from now,
         # and every interval from then on.
@@ -340,6 +375,8 @@
         if not getattr(self, "enabled", False):
             return False
         if not quiet:
+            self.send_changedstate()
+        if not quiet:
             logger.info(u"Disabling client %s", self.name)
         if getattr(self, u"disable_initiator_tag", False):
             gobject.source_remove(self.disable_initiator_tag)
@@ -387,6 +424,9 @@
                                       (self.timeout_milliseconds(),
                                        self.disable))
     
+    def need_approval(self):
+        self.last_approval_request = datetime.datetime.utcnow()
+    
     def start_checker(self):
         """Start a new checker subprocess if one is not running.
         
@@ -421,12 +461,14 @@
                 command = self.checker_command % self.host
             except TypeError:
                 # Escape attributes for the shell
-                escaped_attrs = dict((key,
-                                      re.escape(unicode(str(val),
-                                                        errors=
-                                                        u'replace')))
-                                     for key, val in
-                                     vars(self).iteritems())
+                escaped_attrs = dict(
+                    (attr,
+                     re.escape(unicode(str(getattr(self, attr, u"")),
+                                       errors=
+                                       u'replace')))
+                    for attr in
+                    self.runtime_expansions)
+
                 try:
                     command = self.checker_command % escaped_attrs
                 except TypeError, error:
@@ -478,7 +520,6 @@
                 raise
         self.checker = None
 
-
 def dbus_service_property(dbus_interface, signature=u"v",
                           access=u"readwrite", byte_arrays=False):
     """Decorators for marking methods of a DBusObjectWithProperties to
@@ -675,18 +716,41 @@
     dbus_object_path: dbus.ObjectPath
     bus: dbus.SystemBus()
     """
+    
+    runtime_expansions = (Client.runtime_expansions
+                          + (u"dbus_object_path",))
+    
     # dbus.service.Object doesn't use super(), so we can't either.
     
     def __init__(self, bus = None, *args, **kwargs):
+        self._approvals_pending = 0
         self.bus = bus
         Client.__init__(self, *args, **kwargs)
         # Only now, when this client is initialized, can it show up on
         # the D-Bus
+        client_object_name = unicode(self.name).translate(
+            {ord(u"."): ord(u"_"),
+             ord(u"-"): ord(u"_")})
         self.dbus_object_path = (dbus.ObjectPath
-                                 (u"/clients/"
-                                  + self.name.replace(u".", u"_")))
+                                 (u"/clients/" + client_object_name))
         DBusObjectWithProperties.__init__(self, self.bus,
                                           self.dbus_object_path)
+        
+    def _get_approvals_pending(self):
+        return self._approvals_pending
+    def _set_approvals_pending(self, value):
+        old_value = self._approvals_pending
+        self._approvals_pending = value
+        bval = bool(value)
+        if (hasattr(self, "dbus_object_path")
+            and bval is not bool(old_value)):
+            dbus_bool = dbus.Boolean(bval, variant_level=1)
+            self.PropertyChanged(dbus.String(u"ApprovalPending"),
+                                 dbus_bool)
+
+    approvals_pending = property(_get_approvals_pending,
+                                 _set_approvals_pending)
+    del _get_approvals_pending, _set_approvals_pending
     
     @staticmethod
     def _datetime_to_dbus(dt, variant_level=0):
@@ -699,10 +763,10 @@
         r = Client.enable(self)
         if oldstate != self.enabled:
             # Emit D-Bus signals
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(True, variant_level=1))
             self.PropertyChanged(
-                dbus.String(u"last_enabled"),
+                dbus.String(u"LastEnabled"),
                 self._datetime_to_dbus(self.last_enabled,
                                        variant_level=1))
         return r
@@ -712,7 +776,7 @@
         r = Client.disable(self, quiet=quiet)
         if not quiet and oldstate != self.enabled:
             # Emit D-Bus signal
-            self.PropertyChanged(dbus.String(u"enabled"),
+            self.PropertyChanged(dbus.String(u"Enabled"),
                                  dbus.Boolean(False, variant_level=1))
         return r
     
@@ -730,7 +794,7 @@
         self.checker_callback_tag = None
         self.checker = None
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker_running"),
+        self.PropertyChanged(dbus.String(u"CheckerRunning"),
                              dbus.Boolean(False, variant_level=1))
         if os.WIFEXITED(condition):
             exitstatus = os.WEXITSTATUS(condition)
@@ -751,11 +815,20 @@
         r = Client.checked_ok(self, *args, **kwargs)
         # Emit D-Bus signal
         self.PropertyChanged(
-            dbus.String(u"last_checked_ok"),
+            dbus.String(u"LastCheckedOK"),
             (self._datetime_to_dbus(self.last_checked_ok,
                                     variant_level=1)))
         return r
     
+    def need_approval(self, *args, **kwargs):
+        r = Client.need_approval(self, *args, **kwargs)
+        # Emit D-Bus signal
+        self.PropertyChanged(
+            dbus.String(u"LastApprovalRequest"),
+            (self._datetime_to_dbus(self.last_approval_request,
+                                    variant_level=1)))
+        return r
+    
     def start_checker(self, *args, **kwargs):
         old_checker = self.checker
         if self.checker is not None:
@@ -769,7 +842,7 @@
             # Emit D-Bus signal
             self.CheckerStarted(self.current_checker_command)
             self.PropertyChanged(
-                dbus.String(u"checker_running"),
+                dbus.String(u"CheckerRunning"),
                 dbus.Boolean(True, variant_level=1))
         return r
     
@@ -778,9 +851,21 @@
         r = Client.stop_checker(self, *args, **kwargs)
         if (old_checker is not None
             and getattr(self, u"checker", None) is None):
-            self.PropertyChanged(dbus.String(u"checker_running"),
+            self.PropertyChanged(dbus.String(u"CheckerRunning"),
                                  dbus.Boolean(False, variant_level=1))
         return r
+
+    def _reset_approved(self):
+        self._approved = None
+        return False
+    
+    def approve(self, value=True):
+        self.send_changedstate()
+        self._approved = value
+        gobject.timeout_add(self._timedelta_to_milliseconds
+                            (self.approval_duration),
+                            self._reset_approved)
+    
     
     ## D-Bus methods, signals & properties
     _interface = u"se.bsnet.fukt.Mandos.Client"
@@ -808,17 +893,31 @@
     # GotSecret - signal
     @dbus.service.signal(_interface)
     def GotSecret(self):
-        "D-Bus signal"
+        """D-Bus signal
+        Is sent after a successful transfer of secret from the Mandos
+        server to mandos-client
+        """
         pass
     
     # Rejected - signal
-    @dbus.service.signal(_interface)
-    def Rejected(self):
+    @dbus.service.signal(_interface, signature=u"s")
+    def Rejected(self, reason):
         "D-Bus signal"
         pass
     
+    # NeedApproval - signal
+    @dbus.service.signal(_interface, signature=u"tb")
+    def NeedApproval(self, timeout, default):
+        "D-Bus signal"
+        return self.need_approval()
+    
     ## Methods
     
+    # Approve - method
+    @dbus.service.method(_interface, in_signature=u"b")
+    def Approve(self, value):
+        self.approve(value)
+    
     # CheckedOK - method
     @dbus.service.method(_interface)
     def CheckedOK(self):
@@ -849,43 +948,82 @@
     
     ## Properties
     
-    # name - property
+    # ApprovalPending - property
+    @dbus_service_property(_interface, signature=u"b", access=u"read")
+    def ApprovalPending_dbus_property(self):
+        return dbus.Boolean(bool(self.approvals_pending))
+    
+    # ApprovedByDefault - property
+    @dbus_service_property(_interface, signature=u"b",
+                           access=u"readwrite")
+    def ApprovedByDefault_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.Boolean(self.approved_by_default)
+        self.approved_by_default = bool(value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovedByDefault"),
+                             dbus.Boolean(value, variant_level=1))
+    
+    # ApprovalDelay - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDelay_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self.approval_delay_milliseconds())
+        self.approval_delay = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDelay"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # ApprovalDuration - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def ApprovalDuration_dbus_property(self, value=None):
+        if value is None:       # get
+            return dbus.UInt64(self._timedelta_to_milliseconds(
+                    self.approval_duration))
+        self.approval_duration = datetime.timedelta(0, 0, 0, value)
+        # Emit D-Bus signal
+        self.PropertyChanged(dbus.String(u"ApprovalDuration"),
+                             dbus.UInt64(value, variant_level=1))
+    
+    # Name - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def name_dbus_property(self):
+    def Name_dbus_property(self):
         return dbus.String(self.name)
     
-    # fingerprint - property
+    # Fingerprint - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def fingerprint_dbus_property(self):
+    def Fingerprint_dbus_property(self):
         return dbus.String(self.fingerprint)
     
-    # host - property
+    # Host - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def host_dbus_property(self, value=None):
+    def Host_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.host)
         self.host = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"host"),
+        self.PropertyChanged(dbus.String(u"Host"),
                              dbus.String(value, variant_level=1))
     
-    # created - property
+    # Created - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def created_dbus_property(self):
+    def Created_dbus_property(self):
         return dbus.String(self._datetime_to_dbus(self.created))
     
-    # last_enabled - property
+    # LastEnabled - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
-    def last_enabled_dbus_property(self):
+    def LastEnabled_dbus_property(self):
         if self.last_enabled is None:
             return dbus.String(u"")
         return dbus.String(self._datetime_to_dbus(self.last_enabled))
     
-    # enabled - property
+    # Enabled - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def enabled_dbus_property(self, value=None):
+    def Enabled_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.enabled)
         if value:
@@ -893,10 +1031,10 @@
         else:
             self.disable()
     
-    # last_checked_ok - property
+    # LastCheckedOK - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def last_checked_ok_dbus_property(self, value=None):
+    def LastCheckedOK_dbus_property(self, value=None):
         if value is not None:
             self.checked_ok()
             return
@@ -905,15 +1043,24 @@
         return dbus.String(self._datetime_to_dbus(self
                                                   .last_checked_ok))
     
-    # timeout - property
+    # LastApprovalRequest - property
+    @dbus_service_property(_interface, signature=u"s", access=u"read")
+    def LastApprovalRequest_dbus_property(self):
+        if self.last_approval_request is None:
+            return dbus.String(u"")
+        return dbus.String(self.
+                           _datetime_to_dbus(self
+                                             .last_approval_request))
+    
+    # Timeout - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def timeout_dbus_property(self, value=None):
+    def Timeout_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.timeout_milliseconds())
         self.timeout = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"timeout"),
+        self.PropertyChanged(dbus.String(u"Timeout"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"disable_initiator_tag", None) is None:
             return
@@ -933,15 +1080,15 @@
             self.disable_initiator_tag = (gobject.timeout_add
                                           (time_to_die, self.disable))
     
-    # interval - property
+    # Interval - property
     @dbus_service_property(_interface, signature=u"t",
                            access=u"readwrite")
-    def interval_dbus_property(self, value=None):
+    def Interval_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.interval_milliseconds())
         self.interval = datetime.timedelta(0, 0, 0, value)
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"interval"),
+        self.PropertyChanged(dbus.String(u"Interval"),
                              dbus.UInt64(value, variant_level=1))
         if getattr(self, u"checker_initiator_tag", None) is None:
             return
@@ -951,22 +1098,22 @@
                                       (value, self.start_checker))
         self.start_checker()    # Start one now, too
 
-    # checker - property
+    # Checker - property
     @dbus_service_property(_interface, signature=u"s",
                            access=u"readwrite")
-    def checker_dbus_property(self, value=None):
+    def Checker_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.String(self.checker_command)
         self.checker_command = value
         # Emit D-Bus signal
-        self.PropertyChanged(dbus.String(u"checker"),
+        self.PropertyChanged(dbus.String(u"Checker"),
                              dbus.String(self.checker_command,
                                          variant_level=1))
     
-    # checker_running - property
+    # CheckerRunning - property
     @dbus_service_property(_interface, signature=u"b",
                            access=u"readwrite")
-    def checker_running_dbus_property(self, value=None):
+    def CheckerRunning_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.Boolean(self.checker is not None)
         if value:
@@ -974,20 +1121,46 @@
         else:
             self.stop_checker()
     
-    # object_path - property
+    # ObjectPath - property
     @dbus_service_property(_interface, signature=u"o", access=u"read")
-    def object_path_dbus_property(self):
+    def ObjectPath_dbus_property(self):
         return self.dbus_object_path # is already a dbus.ObjectPath
     
-    # secret = property
+    # Secret = property
     @dbus_service_property(_interface, signature=u"ay",
                            access=u"write", byte_arrays=True)
-    def secret_dbus_property(self, value):
+    def Secret_dbus_property(self, value):
         self.secret = str(value)
     
     del _interface
 
 
+class ProxyClient(object):
+    def __init__(self, child_pipe, fpr, address):
+        self._pipe = child_pipe
+        self._pipe.send(('init', fpr, address))
+        if not self._pipe.recv():
+            raise KeyError()
+
+    def __getattribute__(self, name):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__getattribute__(name)
+        self._pipe.send(('getattr', name))
+        data = self._pipe.recv()
+        if data[0] == 'data':
+            return data[1]
+        if data[0] == 'function':
+            def func(*args, **kwargs):
+                self._pipe.send(('funcall', name, args, kwargs))
+                return self._pipe.recv()[1]
+            return func
+
+    def __setattr__(self, name, value):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__setattr__(name, value)
+        self._pipe.send(('setattr', name, value))
+
+
 class ClientHandler(socketserver.BaseRequestHandler, object):
     """A class to handle client connections.
     
@@ -995,24 +1168,22 @@
     Note: This will run in its own forked process."""
     
     def handle(self):
-        logger.info(u"TCP connection from: %s",
-                    unicode(self.client_address))
-        logger.debug(u"IPC Pipe FD: %d",
-                     self.server.child_pipe[1].fileno())
-        # Open IPC pipe to parent process
-        with contextlib.nested(self.server.child_pipe[1],
-                               self.server.parent_pipe[0]
-                               ) as (ipc, ipc_return):
+        with contextlib.closing(self.server.child_pipe) as child_pipe:
+            logger.info(u"TCP connection from: %s",
+                        unicode(self.client_address))
+            logger.debug(u"Pipe FD: %d",
+                         self.server.child_pipe.fileno())
+
             session = (gnutls.connection
                        .ClientSession(self.request,
                                       gnutls.connection
                                       .X509Credentials()))
-            
+
             # Note: gnutls.connection.X509Credentials is really a
             # generic GnuTLS certificate credentials object so long as
             # no X.509 keys are added to it.  Therefore, we can use it
             # here despite using OpenPGP certificates.
-            
+
             #priority = u':'.join((u"NONE", u"+VERS-TLS1.1",
             #                      u"+AES-256-CBC", u"+SHA1",
             #                      u"+COMP-NULL", u"+CTYPE-OPENPGP",
@@ -1024,7 +1195,7 @@
             (gnutls.library.functions
              .gnutls_priority_set_direct(session._c_object,
                                          priority, None))
-            
+
             # Start communication using the Mandos protocol
             # Get protocol number
             line = self.request.makefile().readline()
@@ -1035,7 +1206,7 @@
             except (ValueError, IndexError, RuntimeError), error:
                 logger.error(u"Unknown protocol version: %s", error)
                 return
-            
+
             # Start GnuTLS connection
             try:
                 session.handshake()
@@ -1045,6 +1216,8 @@
                 # established.  Just abandon the request.
                 return
             logger.debug(u"Handshake succeeded")
+
+            approval_required = False
             try:
                 try:
                     fpr = self.fingerprint(self.peer_certificate
@@ -1054,48 +1227,92 @@
                     return
                 logger.debug(u"Fingerprint: %s", fpr)
 
-                for c in self.server.clients:
-                    if c.fingerprint == fpr:
-                        client = c
+                try:
+                    client = ProxyClient(child_pipe, fpr,
+                                         self.client_address)
+                except KeyError:
+                    return
+                
+                if client.approval_delay:
+                    delay = client.approval_delay
+                    client.approvals_pending += 1
+                    approval_required = True
+                
+                while True:
+                    if not client.enabled:
+                        logger.warning(u"Client %s is disabled",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Disabled")                    
+                        return
+                    
+                    if client._approved or not client.approval_delay:
+                        #We are approved or approval is disabled
                         break
-                else:
-                    ipc.write(u"NOTFOUND %s %s\n"
-                              % (fpr, unicode(self.client_address)))
-                    return
-                
-                class ClientProxy(object):
-                    """Client proxy object.  Not for calling methods."""
-                    def __init__(self, client):
-                        self.client = client
-                    def __getattr__(self, name):
-                        if name.startswith("ipc_"):
-                            def tempfunc():
-                                ipc.write("%s %s\n" % (name[4:].upper(),
-                                                       self.client.name))
-                            return tempfunc
-                        if not hasattr(self.client, name):
-                            raise AttributeError
-                        ipc.write(u"GETATTR %s %s\n"
-                                  % (name, self.client.fingerprint))
-                        return pickle.load(ipc_return)
-                clientproxy = ClientProxy(client)
-                # Have to check if client.enabled, since it is
-                # possible that the client was disabled since the
-                # GnuTLS session was established.
-                if not clientproxy.enabled:
-                    clientproxy.ipc_disabled()
-                    return
-                
-                clientproxy.ipc_sending()
+                    elif client._approved is None:
+                        logger.info(u"Client %s needs approval",
+                                    client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.NeedApproval(
+                                client.approval_delay_milliseconds(),
+                                client.approved_by_default)
+                    else:
+                        logger.warning(u"Client %s was not approved",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Denied")
+                        return
+                    
+                    #wait until timeout or approved
+                    #x = float(client._timedelta_to_milliseconds(delay))
+                    time = datetime.datetime.now()
+                    client.changedstate.acquire()
+                    client.changedstate.wait(float(client._timedelta_to_milliseconds(delay) / 1000))
+                    client.changedstate.release()
+                    time2 = datetime.datetime.now()
+                    if (time2 - time) >= delay:
+                        if not client.approved_by_default:
+                            logger.warning("Client %s timed out while"
+                                           " waiting for approval",
+                                           client.name)
+                            if self.server.use_dbus:
+                                # Emit D-Bus signal
+                                client.Rejected("Approval timed out")
+                            return
+                        else:
+                            break
+                    else:
+                        delay -= time2 - time
+                
                 sent_size = 0
                 while sent_size < len(client.secret):
-                    sent = session.send(client.secret[sent_size:])
+                    try:
+                        sent = session.send(client.secret[sent_size:])
+                    except (gnutls.errors.GNUTLSError), error:
+                        logger.warning("gnutls send failed")
+                        return
                     logger.debug(u"Sent: %d, remaining: %d",
                                  sent, len(client.secret)
                                  - (sent_size + sent))
                     sent_size += sent
+
+                logger.info(u"Sending secret to %s", client.name)
+                # bump the timeout as if seen
+                client.checked_ok()
+                if self.server.use_dbus:
+                    # Emit D-Bus signal
+                    client.GotSecret()
+            
             finally:
-                session.bye()
+                if approval_required:
+                    client.approvals_pending -= 1
+                try:
+                    session.bye()
+                except (gnutls.errors.GNUTLSError), error:
+                    logger.warning("GnuTLS bye failed")
     
     @staticmethod
     def peer_certificate(session):
@@ -1161,30 +1378,39 @@
         return hex_fpr
 
 
-class ForkingMixInWithPipes(socketserver.ForkingMixIn, object):
-    """Like socketserver.ForkingMixIn, but also pass a pipe pair."""
+class MultiprocessingMixIn(object):
+    """Like socketserver.ThreadingMixIn, but with multiprocessing"""
+    def sub_process_main(self, request, address):
+        try:
+            self.finish_request(request, address)
+        except:
+            self.handle_error(request, address)
+        self.close_request(request)
+            
+    def process_request(self, request, address):
+        """Start a new process to process the request."""
+        multiprocessing.Process(target = self.sub_process_main,
+                                args = (request, address)).start()
+
+class MultiprocessingMixInWithPipe(MultiprocessingMixIn, object):
+    """ adds a pipe to the MixIn """
     def process_request(self, request, client_address):
         """Overrides and wraps the original process_request().
         
         This function creates a new pipe in self.pipe
         """
-        # Child writes to child_pipe
-        self.child_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        # Parent writes to parent_pipe
-        self.parent_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        super(ForkingMixInWithPipes,
+        parent_pipe, self.child_pipe = multiprocessing.Pipe()
+
+        super(MultiprocessingMixInWithPipe,
               self).process_request(request, client_address)
-        # Close unused ends for parent
-        self.parent_pipe[0].close() # close read end
-        self.child_pipe[1].close()  # close write end
-        self.add_pipe_fds(self.child_pipe[0], self.parent_pipe[1])
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+        self.child_pipe.close()
+        self.add_pipe(parent_pipe)
+
+    def add_pipe(self, parent_pipe):
         """Dummy function; override as necessary"""
-        child_pipe_fd.close()
-        parent_pipe_fd.close()
-
-
-class IPv6_TCPServer(ForkingMixInWithPipes,
+        pass
+
+class IPv6_TCPServer(MultiprocessingMixInWithPipe,
                      socketserver.TCPServer, object):
     """IPv6-capable TCP server.  Accepts 'None' as address and/or port
     
@@ -1275,14 +1501,15 @@
             return socketserver.TCPServer.server_activate(self)
     def enable(self):
         self.enabled = True
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+    def add_pipe(self, parent_pipe):
         # Call "handle_ipc" for both data and EOF events
-        gobject.io_add_watch(child_pipe_fd.fileno(),
+        gobject.io_add_watch(parent_pipe.fileno(),
                              gobject.IO_IN | gobject.IO_HUP,
                              functools.partial(self.handle_ipc,
-                                               reply = parent_pipe_fd,
-                                               sender= child_pipe_fd))
-    def handle_ipc(self, source, condition, reply=None, sender=None):
+                                               parent_pipe = parent_pipe))
+        
+    def handle_ipc(self, source, condition, parent_pipe=None,
+                   client_object=None):
         condition_names = {
             gobject.IO_IN: u"IN",   # There is data to read.
             gobject.IO_OUT: u"OUT", # Data can be written (without
@@ -1297,70 +1524,58 @@
                                        for cond, name in
                                        condition_names.iteritems()
                                        if cond & condition)
-        logger.debug(u"Handling IPC: FD = %d, condition = %s", source,
-                     conditions_string)
-        
-        # Read a line from the file object
-        cmdline = sender.readline()
-        if not cmdline:             # Empty line means end of file
-            # close the IPC pipes
-            sender.close()
-            reply.close()
-            
-            # Stop calling this function
-            return False
-        
-        logger.debug(u"IPC command: %r", cmdline)
-        
-        # Parse and act on command
-        cmd, args = cmdline.rstrip(u"\r\n").split(None, 1)
-        
-        if cmd == u"NOTFOUND":
-            fpr, address = args.split(None, 1)
-            logger.warning(u"Client not found for fingerprint: %s, ad"
-                           u"dress: %s", fpr, address)
-            if self.use_dbus:
-                # Emit D-Bus signal
-                mandos_dbus_service.ClientNotFound(fpr, address)
-        elif cmd == u"DISABLED":
-            for client in self.clients:
-                if client.name == args:
-                    logger.warning(u"Client %s is disabled", args)
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.Rejected()
-                    break
-            else:
-                logger.error(u"Unknown client %s is disabled", args)
-        elif cmd == u"SENDING":
-            for client in self.clients:
-                if client.name == args:
-                    logger.info(u"Sending secret to %s", client.name)
-                    client.checked_ok()
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.GotSecret()
-                    break
-            else:
-                logger.error(u"Sending secret to unknown client %s",
-                             args)
-        elif cmd == u"GETATTR":
-            attr_name, fpr = args.split(None, 1)
-            for client in self.clients:
-                if client.fingerprint == fpr:
-                    attr_value = getattr(client, attr_name, None)
-                    logger.debug("IPC reply: %r", attr_value)
-                    pickle.dump(attr_value, reply)
-                    break
-            else:
-                logger.error(u"Client %s on address %s requesting "
-                             u"attribute %s not found", fpr, address,
-                             attr_name)
-                pickle.dump(None, reply)
-        else:
-            logger.error(u"Unknown IPC command: %r", cmdline)
-        
-        # Keep calling this function
+        # error or the other end of multiprocessing.Pipe has closed
+        if condition & (gobject.IO_ERR | condition & gobject.IO_HUP):
+            return False
+        
+        # Read a request from the child
+        request = parent_pipe.recv()
+        command = request[0]
+        
+        if command == 'init':
+            fpr = request[1]
+            address = request[2]
+            
+            for c in self.clients:
+                if c.fingerprint == fpr:
+                    client = c
+                    break
+            else:
+                logger.warning(u"Client not found for fingerprint: %s, ad"
+                               u"dress: %s", fpr, address)
+                if self.use_dbus:
+                    # Emit D-Bus signal
+                    mandos_dbus_service.ClientNotFound(fpr, address[0])
+                parent_pipe.send(False)
+                return False
+            
+            gobject.io_add_watch(parent_pipe.fileno(),
+                                 gobject.IO_IN | gobject.IO_HUP,
+                                 functools.partial(self.handle_ipc,
+                                                   parent_pipe = parent_pipe,
+                                                   client_object = client))
+            parent_pipe.send(True)
+            # remove the old hook in favor of the new above hook on same fileno
+            return False
+        if command == 'funcall':
+            funcname = request[1]
+            args = request[2]
+            kwargs = request[3]
+            
+            parent_pipe.send(('data', getattr(client_object, funcname)(*args, **kwargs)))
+
+        if command == 'getattr':
+            attrname = request[1]
+            if callable(client_object.__getattribute__(attrname)):
+                parent_pipe.send(('function',))
+            else:
+                parent_pipe.send(('data', client_object.__getattribute__(attrname)))
+        
+        if command == 'setattr':
+            attrname = request[1]
+            value = request[2]
+            setattr(client_object, attrname, value)
+
         return True
 
 
@@ -1469,6 +1684,8 @@
     parser.add_option("--debug", action=u"store_true",
                       help=u"Debug mode; run in foreground and log to"
                       u" terminal")
+    parser.add_option("--debuglevel", type=u"string", metavar="Level",
+                      help=u"Debug level for stdout output")
     parser.add_option("--priority", type=u"string", help=u"GnuTLS"
                       u" priority string (see GnuTLS documentation)")
     parser.add_option("--servicename", type=u"string",
@@ -1499,6 +1716,7 @@
                         u"servicename": u"Mandos",
                         u"use_dbus": u"True",
                         u"use_ipv6": u"True",
+                        u"debuglevel": u"",
                         }
     
     # Parse config file for server-global settings
@@ -1521,7 +1739,7 @@
     # options, if set.
     for option in (u"interface", u"address", u"port", u"debug",
                    u"priority", u"servicename", u"configdir",
-                   u"use_dbus", u"use_ipv6"):
+                   u"use_dbus", u"use_ipv6", u"debuglevel"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -1536,13 +1754,10 @@
     
     # For convenience
     debug = server_settings[u"debug"]
+    debuglevel = server_settings[u"debuglevel"]
     use_dbus = server_settings[u"use_dbus"]
     use_ipv6 = server_settings[u"use_ipv6"]
-    
-    if not debug:
-        syslogger.setLevel(logging.WARNING)
-        console.setLevel(logging.WARNING)
-    
+
     if server_settings[u"servicename"] != u"Mandos":
         syslogger.setFormatter(logging.Formatter
                                (u'Mandos (%s) [%%(process)d]:'
@@ -1554,6 +1769,8 @@
                         u"interval": u"5m",
                         u"checker": u"fping -q -- %%(host)s",
                         u"host": u"",
+                        u"approval_delay": u"0s",
+                        u"approval_duration": u"1s",
                         }
     client_config = configparser.SafeConfigParser(client_defaults)
     client_config.read(os.path.join(server_settings[u"configdir"],
@@ -1571,11 +1788,12 @@
                               gnutls_priority=
                               server_settings[u"priority"],
                               use_dbus=use_dbus)
-    pidfilename = u"/var/run/mandos.pid"
-    try:
-        pidfile = open(pidfilename, u"w")
-    except IOError:
-        logger.error(u"Could not open file %r", pidfilename)
+    if not debug:
+        pidfilename = u"/var/run/mandos.pid"
+        try:
+            pidfile = open(pidfilename, u"w")
+        except IOError:
+            logger.error(u"Could not open file %r", pidfilename)
     
     try:
         uid = pwd.getpwnam(u"_mandos").pw_uid
@@ -1598,8 +1816,17 @@
         if error[0] != errno.EPERM:
             raise error
     
-    # Enable all possible GnuTLS debugging
+    if not debug and not debuglevel:
+        syslogger.setLevel(logging.WARNING)
+        console.setLevel(logging.WARNING)
+    if debuglevel:
+        level = getattr(logging, debuglevel.upper())
+        syslogger.setLevel(level)
+        console.setLevel(level)
+
     if debug:
+        # Enable all possible GnuTLS debugging
+        
         # "Use a log level over 10 to enable all debugging options."
         # - GnuTLS manual
         gnutls.library.functions.gnutls_global_set_log_level(11)
@@ -1610,6 +1837,16 @@
         
         (gnutls.library.functions
          .gnutls_global_set_log_function(debug_gnutls))
+        
+        # Redirect stdin so all checkers get /dev/null
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
+        os.dup2(null, sys.stdin.fileno())
+        if null > 2:
+            os.close(null)
+    else:
+        # No console logging
+        logger.removeHandler(console)
+    
     
     global main_loop
     # From the Avahi example code
@@ -1633,44 +1870,53 @@
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings[u"interface"])))
+
+    if not debug:
+        # Close all input and output, do double fork, etc.
+        daemon()
+        
+    global multiprocessing_manager
+    multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
         client_class = functools.partial(ClientDBus, bus = bus)
+    def client_config_items(config, section):
+        special_settings = {
+            "approved_by_default":
+                lambda: config.getboolean(section,
+                                          "approved_by_default"),
+            }
+        for name, value in config.items(section):
+            try:
+                yield (name, special_settings[name]())
+            except KeyError:
+                yield (name, value)
+    
     tcp_server.clients.update(set(
             client_class(name = section,
-                         config= dict(client_config.items(section)))
+                         config= dict(client_config_items(
+                        client_config, section)))
             for section in client_config.sections()))
     if not tcp_server.clients:
         logger.warning(u"No clients defined")
-    
-    if debug:
-        # Redirect stdin so all checkers get /dev/null
-        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
-        os.dup2(null, sys.stdin.fileno())
-        if null > 2:
-            os.close(null)
-    else:
-        # No console logging
-        logger.removeHandler(console)
-        # Close all input and output, do double fork, etc.
-        daemon()
-    
-    try:
-        with pidfile:
-            pid = os.getpid()
-            pidfile.write(str(pid) + "\n")
-        del pidfile
-    except IOError:
-        logger.error(u"Could not write to file %r with PID %d",
-                     pidfilename, pid)
-    except NameError:
-        # "pidfile" was never created
-        pass
-    del pidfilename
-    
+        
     if not debug:
+        try:
+            with pidfile:
+                pid = os.getpid()
+                pidfile.write(str(pid) + "\n")
+            del pidfile
+        except IOError:
+            logger.error(u"Could not write to file %r with PID %d",
+                         pidfilename, pid)
+        except NameError:
+            # "pidfile" was never created
+            pass
+        del pidfilename
+        
         signal.signal(signal.SIGINT, signal.SIG_IGN)
+
     signal.signal(signal.SIGHUP, lambda signum, frame: sys.exit())
     signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())
     

=== modified file 'mandos-clients.conf.xml'
--- mandos-clients.conf.xml	2009-12-25 23:13:47 +0000
+++ mandos-clients.conf.xml	2010-09-25 19:55:17 +0000
@@ -3,7 +3,7 @@
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY CONFNAME "mandos-clients.conf">
 <!ENTITY CONFPATH "<filename>/etc/mandos/clients.conf</filename>">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-25">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -100,53 +100,53 @@
     <variablelist>
       
       <varlistentry>
-	<term><option>timeout<literal> = </literal><replaceable
-	>TIME</replaceable></option></term>
-	<listitem>
-	  <para>
-	    This option is <emphasis>optional</emphasis>.
-	  </para>
-	  <para>
-	    The timeout is how long the server will wait (for either a
-	    successful checker run or a client receiving its secret)
-	    until a client is disabled and not allowed to get the data
-	    this server holds.  By default Mandos will use 1 hour.
-	  </para>
-	  <para>
-	    The <replaceable>TIME</replaceable> is specified as a
-	    space-separated number of values, each of which is a
-	    number and a one-character suffix.  The suffix must be one
-	    of <quote>d</quote>, <quote>s</quote>, <quote>m</quote>,
-	    <quote>h</quote>, and <quote>w</quote> for days, seconds,
-	    minutes, hours, and weeks, respectively.  The values are
-	    added together to give the total time value, so all of
-	    <quote><literal>330s</literal></quote>,
-	    <quote><literal>110s 110s 110s</literal></quote>, and
-	    <quote><literal>5m 30s</literal></quote> will give a value
-	    of five minutes and thirty seconds.
-	  </para>
-	</listitem>
-      </varlistentry>
-      
-      <varlistentry>
-	<term><option>interval<literal> = </literal><replaceable
-	>TIME</replaceable></option></term>
-	<listitem>
-	  <para>
-	    This option is <emphasis>optional</emphasis>.
-	  </para>
-	  <para>
-	    How often to run the checker to confirm that a client is
-	    still up.  <emphasis>Note:</emphasis> a new checker will
-	    not be started if an old one is still running.  The server
-	    will wait for a checker to complete until the above
-	    <quote><varname>timeout</varname></quote> occurs, at which
-	    time the client will be disabled, and any running checker
-	    killed.  The default interval is 5 minutes.
-	  </para>
-	  <para>
-	    The format of <replaceable>TIME</replaceable> is the same
-	    as for <varname>timeout</varname> above.
+	<term><option>approval_delay<literal> = </literal><replaceable
+	>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long to wait for external approval before resorting to
+	    use the <option>approved_by_default</option> value.  The
+	    default is <quote>0s</quote>, i.e. not to wait.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> below.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approval_duration<literal> = </literal
+        ><replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How long an external approval lasts.  The default is 1
+	    second.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> below.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>approved_by_default<literal> = </literal
+          >{ <literal >1</literal> | <literal>yes</literal> | <literal
+	  >true</literal> | <literal>on</literal> | <literal
+	  >0</literal> | <literal>no</literal> | <literal
+	  >false</literal> | <literal>off</literal> }</option></term>
+	<listitem>
+	  <para>
+	    Whether to approve a client by default after
+	    the <option>approval_delay</option>.  The default
+	    is <quote>True</quote>.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -196,6 +196,70 @@
       </varlistentry>
       
       <varlistentry>
+	<term><option><literal>host = </literal><replaceable
+	>STRING</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>, but highly
+	    <emphasis>recommended</emphasis> unless the
+	    <option>checker</option> option is modified to a
+	    non-standard value without <quote>%%(host)s</quote> in it.
+	  </para>
+	  <para>
+	    Host name for this client.  This is not used by the server
+	    directly, but can be, and is by default, used by the
+	    checker.  See the <option>checker</option> option.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>interval<literal> = </literal><replaceable
+	>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    How often to run the checker to confirm that a client is
+	    still up.  <emphasis>Note:</emphasis> a new checker will
+	    not be started if an old one is still running.  The server
+	    will wait for a checker to complete until the below
+	    <quote><varname>timeout</varname></quote> occurs, at which
+	    time the client will be disabled, and any running checker
+	    killed.  The default interval is 5 minutes.
+	  </para>
+	  <para>
+	    The format of <replaceable>TIME</replaceable> is the same
+	    as for <varname>timeout</varname> below.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>secfile<literal> = </literal><replaceable
+	>FILENAME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is only used if <option>secret</option> is not
+	    specified, in which case this option is
+	    <emphasis>required</emphasis>.
+	  </para>
+	  <para>
+	    Similar to the <option>secret</option>, except the secret
+	    data is in an external file.  The contents of the file
+	    should <emphasis>not</emphasis> be base64-encoded, but
+	    will be sent to clients verbatim.
+	  </para>
+	  <para>
+	    File names of the form <filename>~user/foo/bar</filename>
+	    and <filename>$<envar>ENVVAR</envar>/foo/bar</filename>
+	    are supported.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
 	<term><option>secret<literal> = </literal><replaceable
 	>BASE64_ENCODED_DATA</replaceable></option></term>
 	<listitem>
@@ -226,42 +290,30 @@
       </varlistentry>
       
       <varlistentry>
-	<term><option>secfile<literal> = </literal><replaceable
-	>FILENAME</replaceable></option></term>
-	<listitem>
-	  <para>
-	    This option is only used if <option>secret</option> is not
-	    specified, in which case this option is
-	    <emphasis>required</emphasis>.
-	  </para>
-	  <para>
-	    Similar to the <option>secret</option>, except the secret
-	    data is in an external file.  The contents of the file
-	    should <emphasis>not</emphasis> be base64-encoded, but
-	    will be sent to clients verbatim.
-	  </para>
-	  <para>
-	    File names of the form <filename>~user/foo/bar</filename>
-	    and <filename>$<envar>ENVVAR</envar>/foo/bar</filename>
-	    are supported.
-	  </para>
-	</listitem>
-      </varlistentry>
-      
-      <varlistentry>
-	<term><option><literal>host = </literal><replaceable
-	>STRING</replaceable></option></term>
-	<listitem>
-	  <para>
-	    This option is <emphasis>optional</emphasis>, but highly
-	    <emphasis>recommended</emphasis> unless the
-	    <option>checker</option> option is modified to a
-	    non-standard value without <quote>%%(host)s</quote> in it.
-	  </para>
-	  <para>
-	    Host name for this client.  This is not used by the server
-	    directly, but can be, and is by default, used by the
-	    checker.  See the <option>checker</option> option.
+	<term><option>timeout<literal> = </literal><replaceable
+	>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    This option is <emphasis>optional</emphasis>.
+	  </para>
+	  <para>
+	    The timeout is how long the server will wait (for either a
+	    successful checker run or a client receiving its secret)
+	    until a client is disabled and not allowed to get the data
+	    this server holds.  By default Mandos will use 1 hour.
+	  </para>
+	  <para>
+	    The <replaceable>TIME</replaceable> is specified as a
+	    space-separated number of values, each of which is a
+	    number and a one-character suffix.  The suffix must be one
+	    of <quote>d</quote>, <quote>s</quote>, <quote>m</quote>,
+	    <quote>h</quote>, and <quote>w</quote> for days, seconds,
+	    minutes, hours, and weeks, respectively.  The values are
+	    added together to give the total time value, so all of
+	    <quote><literal>330s</literal></quote>,
+	    <quote><literal>110s 110s 110s</literal></quote>, and
+	    <quote><literal>5m 30s</literal></quote> will give a value
+	    of five minutes and thirty seconds.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -379,6 +431,8 @@
 fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
 secfile = /etc/mandos/bar-secret
 timeout = 15m
+approved_by_default = False
+approval_delay = 30s
       </programlisting>
     </informalexample>
   </refsect1>

=== modified file 'mandos-ctl'
--- mandos-ctl	2010-09-09 18:16:14 +0000
+++ mandos-ctl	2010-09-25 23:52:17 +0000
@@ -8,34 +8,34 @@
 import locale
 import datetime
 import re
+import os
 
 locale.setlocale(locale.LC_ALL, u'')
 
 tablewords = {
-    'name': u'Name',
-    'enabled': u'Enabled',
-    'timeout': u'Timeout',
-    'last_checked_ok': u'Last Successful Check',
-    'created': u'Created',
-    'interval': u'Interval',
-    'host': u'Host',
-    'fingerprint': u'Fingerprint',
-    'checker_running': u'Check Is Running',
-    'last_enabled': u'Last Enabled',
-    'checker': u'Checker',
+    'Name': u'Name',
+    'Enabled': u'Enabled',
+    'Timeout': u'Timeout',
+    'LastCheckedOK': u'Last Successful Check',
+    'Created': u'Created',
+    'Interval': u'Interval',
+    'Host': u'Host',
+    'Fingerprint': u'Fingerprint',
+    'CheckerRunning': u'Check Is Running',
+    'LastEnabled': u'Last Enabled',
+    'ApprovalPending': u'Approval Is Pending',
+    'ApprovedByDefault': u'Approved By Default',
+    'ApprovalDelay': u"Approval Delay",
+    'ApprovalDuration': u"Approval Duration",
+    'Checker': u'Checker',
     }
-defaultkeywords = ('name', 'enabled', 'timeout', 'last_checked_ok')
+defaultkeywords = ('Name', 'Enabled', 'Timeout', 'LastCheckedOK')
 domain = 'se.bsnet.fukt'
 busname = domain + '.Mandos'
 server_path = '/'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
 version = "1.0.14"
-bus = dbus.SystemBus()
-mandos_dbus_objc = bus.get_object(busname, server_path)
-mandos_serv = dbus.Interface(mandos_dbus_objc,
-                             dbus_interface = server_interface)
-mandos_clients = mandos_serv.GetAllClientsWithProperties()
 
 def timedelta_to_milliseconds(td):
     "Convert a datetime.timedelta object to milliseconds"
@@ -93,18 +93,20 @@
         timevalue += delta
     return timevalue
 
-def print_clients(clients):
+def print_clients(clients, keywords):
     def valuetostring(value, keyword):
         if type(value) is dbus.Boolean:
             return u"Yes" if value else u"No"
-        if keyword in (u"timeout", u"interval"):
+        if keyword in (u"Timeout", u"Interval", u"ApprovalDelay",
+                       u"ApprovalDuration"):
             return milliseconds_to_string(value)
         return unicode(value)
     
     # Create format string to print table rows
     format_string = u' '.join(u'%%-%ds' %
                               max(len(tablewords[key]),
-                                  max(len(valuetostring(client[key], key))
+                                  max(len(valuetostring(client[key],
+                                                        key))
                                       for client in
                                       clients))
                               for key in keywords)
@@ -114,91 +116,199 @@
         print format_string % tuple(valuetostring(client[key], key)
                                     for key in keywords)
 
-parser = OptionParser(version = "%%prog %s" % version)
-parser.add_option("-a", "--all", action="store_true",
-                  help="Print all fields")
-parser.add_option("-e", "--enable", action="store_true",
-                  help="Enable client")
-parser.add_option("-d", "--disable", action="store_true",
-                  help="disable client")
-parser.add_option("-b", "--bump-timeout", action="store_true",
-                  help="Bump timeout for client")
-parser.add_option("--start-checker", action="store_true",
-                  help="Start checker for client")
-parser.add_option("--stop-checker", action="store_true",
-                  help="Stop checker for client")
-parser.add_option("-V", "--is-enabled", action="store_true",
-                  help="Check if client is enabled")
-parser.add_option("-r", "--remove", action="store_true",
-                  help="Remove client")
-parser.add_option("-c", "--checker", type="string",
-                  help="Set checker command for client")
-parser.add_option("-t", "--timeout", type="string",
-                  help="Set timeout for client")
-parser.add_option("-i", "--interval", type="string",
-                  help="Set checker interval for client")
-parser.add_option("-H", "--host", type="string",
-                  help="Set host for client")
-parser.add_option("-s", "--secret", type="string",
-                  help="Set password blob (file) for client")
-options, client_names = parser.parse_args()
-
-# Compile list of clients to process
-clients=[]
-for name in client_names:
-    for path, client in mandos_clients.iteritems():
-        if client['name'] == name:
-            client_objc = bus.get_object(busname, path)
-            clients.append(client_objc)
-            break
-    else:
-        print >> sys.stderr, "Client not found on server: %r" % name
-        sys.exit(1)
-
-if not clients and mandos_clients.values():
-    keywords = defaultkeywords
-    if options.all:
-        keywords = ('name', 'enabled', 'timeout', 'last_checked_ok',
-                    'created', 'interval', 'host', 'fingerprint',
-                    'checker_running', 'last_enabled', 'checker')
-    print_clients(mandos_clients.values())
-
-# Process each client in the list by all selected options
-for client in clients:
-    if options.remove:
-        mandos_serv.RemoveClient(client.__dbus_object_path__)
-    if options.enable:
-        client.Enable(dbus_interface=client_interface)
-    if options.disable:
-        client.Disable(dbus_interface=client_interface)
-    if options.bump_timeout:
-        client.CheckedOK(dbus_interface=client_interface)
-    if options.start_checker:
-        client.StartChecker(dbus_interface=client_interface)
-    if options.stop_checker:
-        client.StopChecker(dbus_interface=client_interface)
-    if options.is_enabled:
-        sys.exit(0 if client.Get(client_interface,
-                                 u"enabled",
-                                 dbus_interface=dbus.PROPERTIES_IFACE)
-                 else 1)
-    if options.checker:
-        client.Set(client_interface, u"checker", options.checker,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.host:
-        client.Set(client_interface, u"host", options.host,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.interval:
-        client.Set(client_interface, u"interval",
-                   timedelta_to_milliseconds
-                   (string_to_delta(options.interval)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.timeout:
-        client.Set(client_interface, u"timeout",
-                   timedelta_to_milliseconds(string_to_delta
-                                             (options.timeout)),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-    if options.secret:
-        client.Set(client_interface, u"secret",
-                   dbus.ByteArray(open(options.secret, u'rb').read()),
-                   dbus_interface=dbus.PROPERTIES_IFACE)
+def has_actions(options):
+    return any((options.enable,
+                options.disable,
+                options.bump_timeout,
+                options.start_checker,
+                options.stop_checker,
+                options.is_enabled,
+                options.remove,
+                options.checker is not None,
+                options.timeout is not None,
+                options.interval is not None,
+                options.approved_by_default is not None,
+                options.approval_delay is not None,
+                options.approval_duration is not None,
+                options.host is not None,
+                options.secret is not None,
+                options.approve,
+                options.deny))
+        
+def main():
+	parser = OptionParser(version = "%%prog %s" % version)
+	parser.add_option("-a", "--all", action="store_true",
+	                  help="Select all clients")
+	parser.add_option("-v", "--verbose", action="store_true",
+	                  help="Print all fields")
+	parser.add_option("-e", "--enable", action="store_true",
+	                  help="Enable client")
+	parser.add_option("-d", "--disable", action="store_true",
+	                  help="disable client")
+	parser.add_option("-b", "--bump-timeout", action="store_true",
+	                  help="Bump timeout for client")
+	parser.add_option("--start-checker", action="store_true",
+	                  help="Start checker for client")
+	parser.add_option("--stop-checker", action="store_true",
+	                  help="Stop checker for client")
+	parser.add_option("-V", "--is-enabled", action="store_true",
+	                  help="Check if client is enabled")
+	parser.add_option("-r", "--remove", action="store_true",
+	                  help="Remove client")
+	parser.add_option("-c", "--checker", type="string",
+	                  help="Set checker command for client")
+	parser.add_option("-t", "--timeout", type="string",
+	                  help="Set timeout for client")
+	parser.add_option("-i", "--interval", type="string",
+	                  help="Set checker interval for client")
+        parser.add_option("--approve-by-default", action="store_true",
+                          dest=u"approved_by_default",
+                          help="Set client to be approved by default")
+        parser.add_option("--deny-by-default", action="store_false",
+                          dest=u"approved_by_default",
+                          help="Set client to be denied by default")
+        parser.add_option("--approval-delay", type="string",
+                          help="Set delay before client approve/deny")
+        parser.add_option("--approval-duration", type="string",
+                          help="Set duration of one client approval")
+	parser.add_option("-H", "--host", type="string",
+	                  help="Set host for client")
+	parser.add_option("-s", "--secret", type="string",
+	                  help="Set password blob (file) for client")
+	parser.add_option("-A", "--approve", action="store_true",
+	                  help="Approve any current client request")
+	parser.add_option("-D", "--deny", action="store_true",
+	                  help="Deny any current client request")
+	options, client_names = parser.parse_args()
+        
+        if has_actions(options) and not client_names and not options.all:
+            parser.error('Options require clients names or --all.')
+        if options.verbose and has_actions(options):
+            parser.error('--verbose can only be used alone or with'
+                         ' --all.')
+        if options.all and not has_actions(options):
+            parser.error('--all requires an action.')
+        
+        try:
+            bus = dbus.SystemBus()
+            mandos_dbus_objc = bus.get_object(busname, server_path)
+        except dbus.exceptions.DBusException:
+            print >> sys.stderr, "Could not connect to Mandos server"
+            sys.exit(1)
+    
+        mandos_serv = dbus.Interface(mandos_dbus_objc,
+                                     dbus_interface = server_interface)
+
+        #block stderr since dbus library prints to stderr
+        null = os.open(os.path.devnull, os.O_RDWR)
+        stderrcopy = os.dup(sys.stderr.fileno())
+        os.dup2(null, sys.stderr.fileno())
+        os.close(null)
+        try:
+            try:
+                mandos_clients = mandos_serv.GetAllClientsWithProperties()
+            finally:
+                #restore stderr
+                os.dup2(stderrcopy, sys.stderr.fileno())
+                os.close(stderrcopy)
+        except dbus.exceptions.DBusException, e:
+            print >> sys.stderr, "Access denied: Accessing mandos server through dbus."
+            sys.exit(1)
+            
+	# Compile dict of (clients: properties) to process
+	clients={}
+        
+        if options.all or not client_names:
+            clients = dict((bus.get_object(busname, path), properties)
+                           for path, properties in
+                           mandos_clients.iteritems())
+        else:
+            for name in client_names:
+                for path, client in mandos_clients.iteritems():
+                    if client['Name'] == name:
+                        client_objc = bus.get_object(busname, path)
+                        clients[client_objc] = client
+                        break
+                else:
+                    print >> sys.stderr, "Client not found on server: %r" % name
+                    sys.exit(1)
+            
+	if not has_actions(options) and clients:
+	    if options.verbose:
+	        keywords = ('Name', 'Enabled', 'Timeout',
+                            'LastCheckedOK', 'Created', 'Interval',
+                            'Host', 'Fingerprint', 'CheckerRunning',
+                            'LastEnabled', 'ApprovalPending',
+                            'ApprovedByDefault', 'ApprovalDelay',
+                            'ApprovalDuration', 'Checker')
+            else:
+                keywords = defaultkeywords
+                
+	    print_clients(clients.values(), keywords)
+        else:
+            # Process each client in the list by all selected options
+            for client in clients:
+                if options.remove:
+                    mandos_serv.RemoveClient(client.__dbus_object_path__)
+                if options.enable:
+                    client.Enable(dbus_interface=client_interface)
+                if options.disable:
+                    client.Disable(dbus_interface=client_interface)
+                if options.bump_timeout:
+                    client.CheckedOK(dbus_interface=client_interface)
+                if options.start_checker:
+                    client.StartChecker(dbus_interface=client_interface)
+                if options.stop_checker:
+                    client.StopChecker(dbus_interface=client_interface)
+                if options.is_enabled:
+                    sys.exit(0 if client.Get(client_interface,
+                                             u"Enabled",
+                                             dbus_interface=dbus.PROPERTIES_IFACE)
+                             else 1)
+                if options.checker:
+                    client.Set(client_interface, u"Checker", options.checker,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.host:
+                    client.Set(client_interface, u"Host", options.host,
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.interval:
+                    client.Set(client_interface, u"Interval",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.interval)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approval_delay:
+                    client.Set(client_interface, u"ApprovalDelay",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.
+                                                approval_delay)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approval_duration:
+                    client.Set(client_interface, u"ApprovalDuration",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.
+                                                approval_duration)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.timeout:
+                    client.Set(client_interface, u"Timeout",
+                               timedelta_to_milliseconds
+                               (string_to_delta(options.timeout)),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.secret:
+                    client.Set(client_interface, u"Secret",
+                               dbus.ByteArray(open(options.secret,
+                                                   u'rb').read()),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approved_by_default is not None:
+                    client.Set(client_interface, u"ApprovedByDefault",
+                               dbus.Boolean(options
+                                            .approved_by_default),
+                               dbus_interface=dbus.PROPERTIES_IFACE)
+                if options.approve:
+                    client.Approve(dbus.Boolean(True),
+                                   dbus_interface=client_interface)
+                elif options.deny:
+                    client.Approve(dbus.Boolean(False),
+                                   dbus_interface=client_interface)
+
+if __name__ == '__main__':
+    main()

=== added file 'mandos-ctl.xml'
--- mandos-ctl.xml	1970-01-01 00:00:00 +0000
+++ mandos-ctl.xml	2010-09-25 23:52:17 +0000
@@ -0,0 +1,570 @@
+<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY COMMANDNAME "mandos-ctl">
+<!ENTITY TIMESTAMP "2010-09-26">
+<!ENTITY % common SYSTEM "common.ent">
+%common;
+]>
+
+<refentry xmlns:xi="http://www.w3.org/2001/XInclude">
+  <refentryinfo>
+    <title>Mandos Manual</title>
+    <!-- NWalsh’s docbook scripts use this to generate the footer: -->
+    <productname>Mandos</productname>
+    <productnumber>&version;</productnumber>
+    <date>&TIMESTAMP;</date>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2010</year>
+      <holder>Teddy Hogeborn</holder>
+      <holder>Björn Påhlsson</holder>
+    </copyright>
+    <xi:include href="legalnotice.xml"/>
+  </refentryinfo>
+  
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      Control the operation of the Mandos server
+    </refpurpose>
+  </refnamediv>
+  
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <group>
+	<arg choice="plain"><option>--enable</option></arg>
+	<arg choice="plain"><option>-e</option></arg>
+	<sbr/>
+	<arg choice="plain"><option>--disable</option></arg>
+	<arg choice="plain"><option>-d</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--bump-timeout</option></arg>
+	<arg choice="plain"><option>-b</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--start-checker</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--stop-checker</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--remove</option></arg>
+	<arg choice="plain"><option>-r</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--checker
+	<replaceable>COMMAND</replaceable></option></arg>
+	<arg choice="plain"><option>-c
+	<replaceable>COMMAND</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--timeout
+	<replaceable>TIME</replaceable></option></arg>
+	<arg choice="plain"><option>-t
+	<replaceable>TIME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--interval
+	<replaceable>TIME</replaceable></option></arg>
+	<arg choice="plain"><option>-i
+	<replaceable>TIME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--approve-by-default</option
+        ></arg>
+	<sbr/>
+	<arg choice="plain"><option>--deny-by-default</option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--approval-delay
+	<replaceable>TIME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--approval-duration
+	<replaceable>TIME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--interval
+	<replaceable>TIME</replaceable></option></arg>
+	<arg choice="plain"><option>-i
+	<replaceable>TIME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--host
+	<replaceable>STRING</replaceable></option></arg>
+	<arg choice="plain"><option>-H
+	<replaceable>STRING</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--secret
+	<replaceable>FILENAME</replaceable></option></arg>
+	<arg choice="plain"><option>-s
+	<replaceable>FILENAME</replaceable></option></arg>
+      </group>
+      <sbr/>
+      <group>
+	<arg choice="plain"><option>--approve</option></arg>
+	<arg choice="plain"><option>-A</option></arg>
+	<sbr/>
+	<arg choice="plain"><option>--deny</option></arg>
+	<arg choice="plain"><option>-D</option></arg>
+      </group>
+      <sbr/>
+      <group choice="req">
+	<arg choice="plain"><option>--all</option></arg>
+	<arg choice="plain"><option>-a</option></arg>
+	<arg rep='repeat' choice='plain'>
+	  <replaceable>CLIENT</replaceable>
+	</arg>
+      </group>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <group>
+	<arg choice="plain"><option>--verbose</option></arg>
+	<arg choice="plain"><option>-v</option></arg>
+      </group>
+      <group>
+	<arg rep='repeat' choice='plain'>
+	  <replaceable>CLIENT</replaceable>
+	</arg>
+      </group>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <group choice="req">
+	<arg choice="plain"><option>--is-enabled</option></arg>
+	<arg choice="plain"><option>-V</option></arg>
+      </group>
+      <arg choice='plain'><replaceable>CLIENT</replaceable></arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <group choice="req">
+	<arg choice="plain"><option>--help</option></arg>
+	<arg choice="plain"><option>-h</option></arg>
+      </group>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <group choice="req">
+	<arg choice="plain"><option>--version</option></arg>
+	<arg choice="plain"><option>-v</option></arg>
+      </group>
+    </cmdsynopsis>
+  </refsynopsisdiv>
+  
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is a program to control the
+      operation of the Mandos server <citerefentry><refentrytitle
+      >mandos</refentrytitle><manvolnum>8</manvolnum></citerefentry>.
+    </para>
+    <para>
+      This program can be used to change client settings, approve or
+      deny client requests, and to remove clients from the server.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="purpose">
+    <title>PURPOSE</title>
+    <para>
+      The purpose of this is to enable <emphasis>remote and unattended
+      rebooting</emphasis> of client host computer with an
+      <emphasis>encrypted root file system</emphasis>.  See <xref
+      linkend="overview"/> for details.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="options">
+    <title>OPTIONS</title>
+    
+    <variablelist>
+      <varlistentry>
+	<term><option>--help</option></term>
+	<term><option>-h</option></term>
+	<listitem>
+	  <para>
+	    Show a help message and exit
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--enable</option></term>
+	<term><option>-e</option></term>
+	<listitem>
+	  <para>
+	    Enable client(s).  An enabled client will be eligble to
+	    receive its secret.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--disable</option></term>
+	<term><option>-d</option></term>
+	<listitem>
+	  <para>
+	    Disable client(s).  A disabled client will not be eligble
+	    to receive its secret, and no checkers will be started for
+	    it.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--bump-timeout</option></term>
+	<listitem>
+	  <para>
+	    Bump the timeout of the specified client(s), just as if a
+	    checker had completed successfully for it/them.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--start-checker</option></term>
+	<listitem>
+	  <para>
+	    Start a new checker now for the specified client(s).
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--stop-checker</option></term>
+	<listitem>
+	  <para>
+	    Stop any running checker for the specified client(s).
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--remove</option></term>
+	<term><option>-r</option></term>
+	<listitem>
+	  <para>
+	    Remove the specified client(s) from the server.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--checker
+	<replaceable>COMMAND</replaceable></option></term>
+	<term><option>-c
+	<replaceable>COMMAND</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>checker</varname> option of the specified
+	    client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--timeout
+	<replaceable>TIME</replaceable></option></term>
+	<term><option>-t
+	<replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>timeout</varname> option of the specified
+	    client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--interval
+	<replaceable>TIME</replaceable></option></term>
+	<term><option>-i
+	<replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>interval</varname> option of the
+	    specified client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--approve-by-default</option></term>
+	<term><option>--deny-by-default</option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>approved_by_default</varname> option of
+	    the specified client(s) to <literal>True</literal> or
+	    <literal>False</literal>, respectively; see
+	    <citerefentry><refentrytitle
+            >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--approval-delay
+	<replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>approval_delay</varname> option of the
+	    specified client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--approval-duration
+	<replaceable>TIME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>approval_duration</varname> option of the
+	    specified client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--host
+	<replaceable>STRING</replaceable></option></term>
+	<term><option>-H
+	<replaceable>STRING</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>host</varname> option of the specified
+	    client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--secret
+	<replaceable>FILENAME</replaceable></option></term>
+	<term><option>-s
+	<replaceable>FILENAME</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Set the <varname>secfile</varname> option of the specified
+	    client(s); see <citerefentry><refentrytitle
+	    >mandos-clients.conf</refentrytitle><manvolnum
+            >5</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--approve</option></term>
+	<term><option>-A</option></term>
+	<listitem>
+	  <para>
+	    Approve client(s) if currently waiting for approval.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--deny</option></term>
+	<term><option>-D</option></term>
+	<listitem>
+	  <para>
+	    Deny client(s) if currently waiting for approval.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--all</option></term>
+	<term><option>-a</option></term>
+	<listitem>
+	  <para>
+	    Make the client-modifying options modify <emphasis
+	    >all</emphasis> clients.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--verbose</option></term>
+	<term><option>-v</option></term>
+	<listitem>
+	  <para>
+	    Show all client settings, not just a subset.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--is-enabled</option></term>
+	<term><option>-V</option></term>
+	<listitem>
+	  <para>
+	    Check if a single client is enabled or not, and exit with
+	    a successful exit status only if the client is enabled.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+    </variablelist>
+  </refsect1>
+  
+  <refsect1 id="overview">
+    <title>OVERVIEW</title>
+    <xi:include href="overview.xml"/>
+    <para>
+      This program is a small utility to generate new OpenPGP keys for
+      new Mandos clients, and to generate sections for inclusion in
+      <filename>clients.conf</filename> on the server.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="exit_status">
+    <title>EXIT STATUS</title>
+    <para>
+      If the <option>--is-enabled</option> option is used, the exit
+      status will be 0 only if the specified client is enabled.
+    </para>
+  </refsect1>
+  
+<!--   <refsect1 id="bugs"> -->
+<!--     <title>BUGS</title> -->
+<!--     <para> -->
+<!--     </para> -->
+<!--   </refsect1> -->
+  
+  <refsect1 id="example">
+    <title>EXAMPLE</title>
+    <informalexample>
+      <para>
+	To list all clients:
+      </para>
+      <para>
+	<userinput>&COMMANDNAME;</userinput>
+      </para>
+    </informalexample>
+    
+    <informalexample>
+      <para>
+	To list <emphasis>all</emphasis> settings for the clients
+        named <quote>foo1.example.org</quote> and <quote
+        >foo2.example.org</quote>:
+      </para>
+      <para>
+
+<!-- do not wrap this line -->
+<userinput>&COMMANDNAME; --verbose foo1.example.org foo2.example.org</userinput>
+
+      </para>
+    </informalexample>
+    
+    <informalexample>
+      <para>
+	To enable all clients:
+      </para>
+      <para>
+	<userinput>&COMMANDNAME; --enable --all</userinput>
+      </para>
+    </informalexample>
+    
+    <informalexample>
+      <para>
+	To change timeout and interval value for the clients
+        named <quote>foo1.example.org</quote> and <quote
+        >foo2.example.org</quote>:
+      </para>
+      <para>
+
+<!-- do not wrap this line -->
+<userinput>&COMMANDNAME; --timeout="5m" --interval="1m" foo1.example.org foo2.example.org</userinput>
+
+      </para>
+    </informalexample>
+    
+    <informalexample>
+      <para>
+	To approve all clients currently waiting for it:
+      </para>
+      <para>
+	<userinput>&COMMANDNAME; --approve --all</userinput>
+      </para>
+    </informalexample>
+  </refsect1>
+  
+  <refsect1 id="security">
+    <title>SECURITY</title>
+    <para>
+      This program must be permitted to access the Mandos server via
+      the D-Bus interface.  This normally requires the root user, but
+      could be configured otherwise by reconfiguring the D-Bus server.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="see_also">
+    <title>SEE ALSO</title>
+    <para>
+      <citerefentry><refentrytitle>mandos</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>mandos-clients.conf</refentrytitle>
+      <manvolnum>5</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>mandos-monitor</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>
+    </para>
+  </refsect1>
+  
+</refentry>
+<!-- Local Variables: -->
+<!-- time-stamp-start: "<!ENTITY TIMESTAMP [\"']" -->
+<!-- time-stamp-end: "[\"']>" -->
+<!-- time-stamp-format: "%:y-%02m-%02d" -->
+<!-- End: -->

=== modified file 'mandos-keygen'
--- mandos-keygen	2010-09-09 18:16:14 +0000
+++ mandos-keygen	2010-09-13 06:19:45 +0000
@@ -293,12 +293,13 @@
 	stty -echo
 	echo -n "Enter passphrase: " >&2
 	first="$(head --lines=1 | tr --delete '\n')"
-	echo -n -e "\nRepeat passphrase: " >&2
+	echo >&2
+	echo -n "Repeat passphrase: " >&2
 	second="$(head --lines=1 | tr --delete '\n')"
 	echo >&2
 	stty echo
 	if [ "$first" != "$second" ]; then
-	    echo -e "Passphrase mismatch" >&2
+	    echo "Passphrase mismatch" >&2
 	    touch "$RINGDIR"/mismatch
 	else
 	    echo -n "$first"

=== modified file 'mandos-monitor'
--- mandos-monitor	2010-09-05 20:19:02 +0000
+++ mandos-monitor	2010-09-26 17:36:30 +0000
@@ -23,11 +23,14 @@
 
 locale.setlocale(locale.LC_ALL, u'')
 
+import logging
+logging.getLogger('dbus.proxies').setLevel(logging.CRITICAL)
+
 # Some useful constants
 domain = 'se.bsnet.fukt'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
-version = "1.0.14"
+version = "1.0.15"
 
 # Always run in monochrome mode
 urwid.curses_display.curses.has_colors = lambda : False
@@ -70,8 +73,10 @@
         self.properties.update(
             self.proxy.GetAll(client_interface,
                               dbus_interface = dbus.PROPERTIES_IFACE))
-        super(MandosClientPropertyCache, self).__init__(
-            proxy_object=proxy_object, *args, **kwargs)
+
+        #XXX This break good super behaviour!
+#        super(MandosClientPropertyCache, self).__init__(
+#            *args, **kwargs)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -97,6 +102,7 @@
         self.logger = logger
         
         self._update_timer_callback_tag = None
+        self._update_timer_callback_lock = 0
         self.last_checker_failed = False
         
         # The widget shown normally
@@ -108,6 +114,25 @@
             *args, **kwargs)
         self.update()
         self.opened = False
+        
+        last_checked_ok = isoformat_to_datetime(self.properties
+                                                [u"LastCheckedOK"])
+        if last_checked_ok is None:
+            self.last_checker_failed = True
+        else:
+            self.last_checker_failed = ((datetime.datetime.utcnow()
+                                         - last_checked_ok)
+                                        > datetime.timedelta
+                                        (milliseconds=
+                                         self.properties
+                                         [u"Interval"]))
+        
+        if self.last_checker_failed:
+            self.using_timer(True)
+        
+        if self.need_approval:
+            self.using_timer(True)
+        
         self.proxy.connect_to_signal(u"CheckerCompleted",
                                      self.checker_completed,
                                      client_interface,
@@ -120,71 +145,93 @@
                                      self.got_secret,
                                      client_interface,
                                      byte_arrays=True)
+        self.proxy.connect_to_signal(u"NeedApproval",
+                                     self.need_approval,
+                                     client_interface,
+                                     byte_arrays=True)
         self.proxy.connect_to_signal(u"Rejected",
                                      self.rejected,
                                      client_interface,
                                      byte_arrays=True)
-        last_checked_ok = isoformat_to_datetime(self.properties
-                                                ["last_checked_ok"])
-        if last_checked_ok is None:
-            self.last_checker_failed = True
+    
+    def property_changed(self, property=None, value=None):
+        super(self, MandosClientWidget).property_changed(property,
+                                                         value)
+        if property == u"ApprovalPending":
+            using_timer(bool(value))
+        
+    def using_timer(self, flag):
+        """Call this method with True or False when timer should be
+        activated or deactivated.
+        """
+        old = self._update_timer_callback_lock
+        if flag:
+            self._update_timer_callback_lock += 1
         else:
-            self.last_checker_failed = ((datetime.datetime.utcnow()
-                                         - last_checked_ok)
-                                        > datetime.timedelta
-                                        (milliseconds=
-                                         self.properties["interval"]))
-        if self.last_checker_failed:
+            self._update_timer_callback_lock -= 1
+        if old == 0 and self._update_timer_callback_lock:
             self._update_timer_callback_tag = (gobject.timeout_add
                                                (1000,
                                                 self.update_timer))
+        elif old and self._update_timer_callback_lock == 0:
+            gobject.source_remove(self._update_timer_callback_tag)
+            self._update_timer_callback_tag = None
     
     def checker_completed(self, exitstatus, condition, command):
         if exitstatus == 0:
             if self.last_checker_failed:
                 self.last_checker_failed = False
-                gobject.source_remove(self._update_timer_callback_tag)
-                self._update_timer_callback_tag = None
-            self.logger(u'Checker for client %s (command "%s")'
-                        u' was successful'
-                        % (self.properties[u"name"], command))
+                self.using_timer(False)
+            #self.logger(u'Checker for client %s (command "%s")'
+            #            u' was successful'
+            #            % (self.properties[u"Name"], command))
             self.update()
             return
         # Checker failed
         if not self.last_checker_failed:
             self.last_checker_failed = True
-            self._update_timer_callback_tag = (gobject.timeout_add
-                                               (1000,
-                                                self.update_timer))
+            self.using_timer(True)
         if os.WIFEXITED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' failed with exit code %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WEXITSTATUS(condition)))
         elif os.WIFSIGNALED(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' was killed by signal %s'
-                        % (self.properties[u"name"], command,
+                        % (self.properties[u"Name"], command,
                            os.WTERMSIG(condition)))
         elif os.WCOREDUMP(condition):
             self.logger(u'Checker for client %s (command "%s")'
                         u' dumped core'
-                        % (self.properties[u"name"], command))
+                        % (self.properties[u"Name"], command))
         else:
-            self.logger(u'Checker for client %s completed mysteriously')
+            self.logger(u'Checker for client %s completed'
+                        u' mysteriously')
         self.update()
     
     def checker_started(self, command):
-        self.logger(u'Client %s started checker "%s"'
-                    % (self.properties[u"name"], unicode(command)))
+        #self.logger(u'Client %s started checker "%s"'
+        #            % (self.properties[u"Name"], unicode(command)))
+        pass
     
     def got_secret(self):
+        self.last_checker_failed = False
         self.logger(u'Client %s received its secret'
-                    % self.properties[u"name"])
-    
-    def rejected(self):
-        self.logger(u'Client %s was rejected'
-                    % self.properties[u"name"])
+                    % self.properties[u"Name"])
+    
+    def need_approval(self, timeout, default):
+        if not default:
+            message = u'Client %s needs approval within %s seconds'
+        else:
+            message = u'Client %s will get its secret in %s seconds'
+        self.logger(message
+                    % (self.properties[u"Name"], timeout/1000))
+        self.using_timer(True)
+    
+    def rejected(self, reason):
+        self.logger(u'Client %s was rejected; reason: %s'
+                    % (self.properties[u"Name"], reason))
     
     def selectable(self):
         """Make this a "selectable" widget.
@@ -212,31 +259,46 @@
                           u"bold-underline-blink":
                               u"bold-underline-blink-standout",
                           }
-        
+
         # Rebuild focus and non-focus widgets using current properties
-        self._text = (u'%(name)s: %(enabled)s%(timer)s'
-                      % { u"name": self.properties[u"name"],
-                          u"enabled":
-                              (u"enabled"
-                               if self.properties[u"enabled"]
-                               else u"DISABLED"),
-                          u"timer": (unicode(datetime.timedelta
-                                             (milliseconds =
-                                              self.properties
-                                              [u"timeout"])
-                                             - (datetime.datetime
-                                                .utcnow()
-                                                - isoformat_to_datetime
-                                                (max((self.properties
-                                                 ["last_checked_ok"]
-                                                 or
-                                                 self.properties
-                                                 ["created"]),
-                                                    self.properties[u"last_enabled"]))))
-                                     if (self.last_checker_failed
-                                         and self.properties
-                                         [u"enabled"])
-                                     else u"")})
+
+        # Base part of a client. Name!
+        base = (u'%(name)s: '
+                      % {u"name": self.properties[u"Name"]})
+        if not self.properties[u"Enabled"]:
+            message = u"DISABLED"
+        elif self.properties[u"ApprovalPending"]:
+            timeout = datetime.timedelta(milliseconds
+                                         = self.properties
+                                         [u"ApprovalDelay"])
+            last_approval_request = isoformat_to_datetime(
+                self.properties[u"LastApprovalRequest"])
+            if last_approval_request is not None:
+                timer = timeout - (datetime.datetime.utcnow()
+                                   - last_approval_request)
+            else:
+                timer = datetime.timedelta()
+            if self.properties[u"ApprovedByDefault"]:
+                message = u"Approval in %s. (d)eny?"
+            else:
+                message = u"Denial in %s. (a)pprove?"
+            message = message % unicode(timer).rsplit(".", 1)[0]
+        elif self.last_checker_failed:
+            timeout = datetime.timedelta(milliseconds
+                                         = self.properties
+                                         [u"Timeout"])
+            last_ok = isoformat_to_datetime(
+                max((self.properties[u"LastCheckedOK"]
+                     or self.properties[u"Created"]),
+                    self.properties[u"LastEnabled"]))
+            timer = timeout - (datetime.datetime.utcnow() - last_ok)
+            message = (u'A checker has failed! Time until client'
+                       u' gets disabled: %s'
+                           % unicode(timer).rsplit(".", 1)[0])
+        else:
+            message = u"enabled"
+        self._text = "%s%s" % (base, message)
+            
         if not urwid.supports_unicode():
             self._text = self._text.encode("ascii", "replace")
         textlist = [(u"normal", self._text)]
@@ -274,19 +336,25 @@
     def keypress(self, (maxcol,), key):
         """Handle keys.
         This overrides the method from urwid.FlowWidget"""
-        if key == u"e" or key == u"+":
-            self.proxy.Enable()
-        elif key == u"d" or key == u"-":
-            self.proxy.Disable()
+        if key == u"+":
+            self.proxy.Enable(dbus_interface = client_interface)
+        elif key == u"-":
+            self.proxy.Disable(dbus_interface = client_interface)
+        elif key == u"a":
+            self.proxy.Approve(dbus.Boolean(True, variant_level=1),
+                               dbus_interface = client_interface)
+        elif key == u"d":
+            self.proxy.Approve(dbus.Boolean(False, variant_level=1),
+                                  dbus_interface = client_interface)
         elif key == u"r" or key == u"_" or key == u"ctrl k":
             self.server_proxy_object.RemoveClient(self.proxy
                                                   .object_path)
         elif key == u"s":
-            self.proxy.StartChecker()
+            self.proxy.StartChecker(dbus_interface = client_interface)
         elif key == u"S":
-            self.proxy.StopChecker()
+            self.proxy.StopChecker(dbus_interface = client_interface)
         elif key == u"C":
-            self.proxy.CheckedOK()
+            self.proxy.CheckedOK(dbus_interface = client_interface)
         # xxx
 #         elif key == u"p" or key == "=":
 #             self.proxy.pause()
@@ -315,7 +383,8 @@
     use them as an excuse to shift focus away from this widget.
     """
     def keypress(self, (maxcol, maxrow), key):
-        ret = super(ConstrainedListBox, self).keypress((maxcol, maxrow), key)
+        ret = super(ConstrainedListBox, self).keypress((maxcol,
+                                                        maxrow), key)
         if ret in (u"up", u"down"):
             return
         return ret
@@ -438,10 +507,13 @@
         Call this when the widget layout needs to change"""
         self.uilist = []
         #self.uilist.append(urwid.ListBox(self.clients))
-        self.uilist.append(urwid.Frame(ConstrainedListBox(self.clients),
+        self.uilist.append(urwid.Frame(ConstrainedListBox(self.
+                                                          clients),
                                        #header=urwid.Divider(),
                                        header=None,
-                                       footer=urwid.Divider(div_char=self.divider)))
+                                       footer=
+                                       urwid.Divider(div_char=
+                                                     self.divider)))
         if self.log_visible:
             self.uilist.append(self.logbox)
             pass
@@ -465,8 +537,8 @@
         """Toggle visibility of the log buffer."""
         self.log_visible = not self.log_visible
         self.rebuild()
-        self.log_message(u"Log visibility changed to: "
-                         + unicode(self.log_visible))
+        #self.log_message(u"Log visibility changed to: "
+        #                 + unicode(self.log_visible))
     
     def change_log_display(self):
         """Change type of log display.
@@ -477,7 +549,7 @@
             self.log_wrap = u"clip"
         for textwidget in self.log:
             textwidget.set_wrap_mode(self.log_wrap)
-        self.log_message(u"Wrap mode: " + self.log_wrap)
+        #self.log_message(u"Wrap mode: " + self.log_wrap)
     
     def find_and_remove_client(self, path, name):
         """Find an client from its object path and remove it.
@@ -510,7 +582,7 @@
         if path is None:
             path = client.proxy.object_path
         self.clients_dict[path] = client
-        self.clients.sort(None, lambda c: c.properties[u"name"])
+        self.clients.sort(None, lambda c: c.properties[u"Name"])
         self.refresh()
     
     def remove_client(self, client, path=None):
@@ -591,12 +663,14 @@
                 self.log_message_raw((u"bold",
                                       u"  "
                                       .join((u"Clients:",
-                                             u"e: Enable",
-                                             u"d: Disable",
+                                             u"+: Enable",
+                                             u"-: Disable",
                                              u"r: Remove",
                                              u"s: Start new checker",
                                              u"S: Stop checker",
-                                             u"C: Checker OK"))))
+                                             u"C: Checker OK",
+                                             u"a: Approve",
+                                             u"d: Deny"))))
                 self.refresh()
             elif key == u"tab":
                 if self.topwidget.get_focus() is self.logbox:
@@ -630,6 +704,8 @@
 ui = UserInterface()
 try:
     ui.run()
+except KeyboardInterrupt:
+    ui.screen.stop()
 except Exception, e:
     ui.log_message(unicode(e))
     ui.screen.stop()

=== added file 'mandos-monitor.xml'
--- mandos-monitor.xml	1970-01-01 00:00:00 +0000
+++ mandos-monitor.xml	2010-09-14 18:22:03 +0000
@@ -0,0 +1,233 @@
+<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY COMMANDNAME "mandos-monitor">
+<!ENTITY TIMESTAMP "2010-09-14">
+<!ENTITY % common SYSTEM "common.ent">
+%common;
+]>
+
+<refentry xmlns:xi="http://www.w3.org/2001/XInclude">
+  <refentryinfo>
+    <title>Mandos Manual</title>
+    <!-- NWalsh’s docbook scripts use this to generate the footer: -->
+    <productname>Mandos</productname>
+    <productnumber>&version;</productnumber>
+    <date>&TIMESTAMP;</date>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2010</year>
+      <holder>Teddy Hogeborn</holder>
+      <holder>Björn Påhlsson</holder>
+    </copyright>
+    <xi:include href="legalnotice.xml"/>
+  </refentryinfo>
+  
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      Text-based GUI to control the Mandos server.
+    </refpurpose>
+  </refnamediv>
+  
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+    </cmdsynopsis>
+  </refsynopsisdiv>
+  
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is an interactive program to
+      monitor and control the operations of the Mandos server (see
+      <citerefentry><refentrytitle>mandos</refentrytitle><manvolnum
+      >8</manvolnum></citerefentry>).
+    </para>
+  </refsect1>
+  
+  <refsect1 id="purpose">
+    <title>PURPOSE</title>
+    <para>
+      The purpose of this is to enable <emphasis>remote and unattended
+      rebooting</emphasis> of client host computer with an
+      <emphasis>encrypted root file system</emphasis>.  See <xref
+      linkend="overview"/> for details.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="overview">
+    <title>OVERVIEW</title>
+    <xi:include href="overview.xml"/>
+    <para>
+      This program is used to monitor and control the Mandos server.
+      In particular, it can be used to approve Mandos clients which
+      have been configured to require approval.  It also shows all
+      significant events reported by the Mandos server.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="keys">
+    <title>KEYS</title>
+    <para>
+      This program is used to monitor and control the Mandos server.
+      In particular, it can be used to approve Mandos clients which
+      have been configured to require approval.  It also shows all
+      significant events reported by the Mandos server.
+    </para>
+    <table>
+      <title>Global Keys</title><tgroup cols="2"><thead>
+      <row>
+	<entry>Keys</entry>
+	<entry>Function</entry>
+      </row>
+      </thead><tbody>
+      <row>
+	<entry>q, Q</entry>
+	<entry>Quit</entry>
+      </row>
+      <row>
+	<entry>Ctrl-L</entry>
+	<entry>Redraw screen</entry>
+      </row>
+      <row>
+	<entry>?, F1</entry>
+	<entry>Show help</entry>
+      </row>
+      <row>
+	<entry>l, D</entry>
+	<entry>Toggle log window</entry>
+      </row>
+      <row>
+	<entry>TAB</entry>
+	<entry>Switch window</entry>
+      </row>
+      <row>
+	<entry>w, i</entry>
+	<entry>Toggle log window line wrap</entry>
+      </row>
+      <row>
+	<entry>Up, Ctrl-P, k</entry>
+	<entry>Move up a line</entry>
+      </row>
+      <row>
+	<entry>Down, Ctrl-N, j</entry>
+	<entry>Move down a line</entry>
+      </row>
+      <row>
+	<entry>PageUp, Meta-V, b</entry>
+	<entry>Move up a page</entry>
+      </row>
+      <row>
+	<entry>PageDown, Ctrl-V, SPACE, f</entry>
+	<entry>Move down a page</entry>
+      </row>
+    </tbody></tgroup></table>
+    <table>
+      <title>Client List Keys</title><tgroup cols="2"><thead>
+      <row>
+	<entry>Keys</entry>
+	<entry>Function</entry>
+      </row>
+      </thead><tbody>
+      <row>
+	<entry>+</entry>
+	<entry>Enable client</entry>
+      </row>
+      <row>
+	<entry>-</entry>
+	<entry>Disable client</entry>
+      </row>
+      <row>
+	<entry>a</entry>
+	<entry>Approve client</entry>
+      </row>
+      <row>
+	<entry>d</entry>
+	<entry>Deny client</entry>
+      </row>
+      <row>
+	<entry>r, _, Ctrl-K</entry>
+	<entry>Remove client</entry>
+      </row>
+      <row>
+	<entry>s</entry>
+	<entry>Start checker for client</entry>
+      </row>
+      <row>
+	<entry>S</entry>
+	<entry>Stop checker for client</entry>
+      </row>
+      <row>
+	<entry>C</entry>
+	<entry>Force a successful check for this client.</entry>
+      </row>
+    </tbody></tgroup></table>
+  </refsect1>
+  
+  <refsect1 id="bugs">
+    <title>BUGS</title>
+    <para>
+      This program can currently only be used to monitor and control a
+      Mandos server with the default D-Bus service name of
+      <quote>Mandos</quote>.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="example">
+    <title>EXAMPLE</title>
+    <informalexample>
+      <para>
+	This program takes no options:
+      </para>
+      <para>
+	<userinput>&COMMANDNAME;</userinput>
+      </para>
+    </informalexample>
+  </refsect1>
+  
+  <refsect1 id="security">
+    <title>SECURITY</title>
+    <para>
+      This program must be permitted to access the Mandos server via
+      the D-Bus interface.  This normally requires the root user, but
+      could be configured otherwise by reconfiguring the D-Bus server.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="see_also">
+    <title>SEE ALSO</title>
+    <para>
+      <citerefentry><refentrytitle>mandos</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>mandos-ctl</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>
+    </para>
+  </refsect1>
+  
+</refentry>
+<!-- Local Variables: -->
+<!-- time-stamp-start: "<!ENTITY TIMESTAMP [\"']" -->
+<!-- time-stamp-end: "[\"']>" -->
+<!-- time-stamp-format: "%:y-%02m-%02d" -->
+<!-- End: -->

=== modified file 'mandos.lsm'
--- mandos.lsm	2010-09-09 18:16:14 +0000
+++ mandos.lsm	2010-09-14 18:22:03 +0000
@@ -1,7 +1,7 @@
 Begin4
 Title:		Mandos
 Version:	1.0.14
-Entered-date:	2009-10-25
+Entered-date:	2010-09-14
 Description:	The Mandos system allows computers to have encrypted
 root file systems and at the same time be capable of remote and/or
 unattended reboots.

=== modified file 'mandos.xml'
--- mandos.xml	2010-09-09 18:16:14 +0000
+++ mandos.xml	2010-09-25 21:24:49 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos">
-<!ENTITY TIMESTAMP "2009-12-09">
+<!ENTITY TIMESTAMP "2010-09-25">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -338,6 +338,26 @@
     </para>
   </refsect1>
   
+  <refsect1 id="approval">
+    <title>APPROVAL</title>
+    <para>
+      The server can be configured to require manual approval for a
+      client before it is sent its secret.  The delay to wait for such
+      approval and the default action (approve or deny) can be
+      configured both globally and per client; see <citerefentry>
+      <refentrytitle>mandos-clients.conf</refentrytitle>
+      <manvolnum>5</manvolnum></citerefentry>.  By default all clients
+      will be approved immediately without delay.
+    </para>
+    <para>
+      This can be used to deny a client its secret if not manually
+      approved within a specified time.  It can also be used to make
+      the server delay before giving a client its secret, allowing
+      optional manual denying of this specific client.
+    </para>
+    
+  </refsect1>
+  
   <refsect1 id="logging">
     <title>LOGGING</title>
     <para>
@@ -353,8 +373,8 @@
     <para>
       The server will by default provide a D-Bus system bus interface.
       This interface will only be accessible by the root user or a
-      Mandos-specific user, if such a user exists.
-      <!-- XXX -->
+      Mandos-specific user, if such a user exists.  For documentation
+      of the D-Bus API, see the file <filename>DBUS-API</filename>.
     </para>
   </refsect1>
   
@@ -418,8 +438,8 @@
 	<term><filename>/var/run/mandos.pid</filename></term>
 	<listitem>
 	  <para>
-	    The file containing the process id of
-	    <command>&COMMANDNAME;</command>.
+	    The file containing the process id of the
+	    <command>&COMMANDNAME;</command> process started last.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -458,11 +478,6 @@
       has some security implications, see <xref linkend="clients"/>.
     </para>
     <para>
-      There is currently no way of querying the server of the current
-      status of clients, other than analyzing its <systemitem
-      class="service">syslog</systemitem> output.
-    </para>
-    <para>
       There is no fine-grained control over logging and debug output.
     </para>
     <para>

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2010-03-27 18:39:02 +0000
+++ plugin-runner.c	2010-09-09 22:06:10 +0000
@@ -28,7 +28,7 @@
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
-#include <stdio.h>		/* perror, fileno(), fprintf(),
+#include <stdio.h>		/* fileno(), fprintf(),
 				   stderr, STDOUT_FILENO */
 #include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
 				   stat, waitpid(), WIFEXITED(),
@@ -70,6 +70,8 @@
 #include <inttypes.h>		/* intmax_t, PRIdMAX, strtoimax() */
 #include <sysexits.h>		/* EX_OSERR, EX_USAGE, EX_IOERR,
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
+#include <errno.h> 		/* errno */
+#include <error.h>		/* error() */
 
 #define BUFFER_SIZE 256
 
@@ -266,7 +268,7 @@
 	/* No child processes */
 	break;
       }
-      perror("waitpid");
+      error(0, errno, "waitpid");
     }
     
     /* A child exited, find it in process_list */
@@ -357,13 +359,13 @@
   sigemptyset(&sigchld_action.sa_mask);
   ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     exitstatus = EX_OSERR;
     goto fallback;
   }
   ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
     goto fallback;
   }
@@ -599,7 +601,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -626,7 +628,7 @@
     custom_argc = 1;
     custom_argv = malloc(sizeof(char*) * 2);
     if(custom_argv == NULL){
-      perror("malloc");
+      error(0, errno, "malloc");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -649,7 +651,7 @@
 	}
 	new_arg = strdup(p);
 	if(new_arg == NULL){
-	  perror("strdup");
+	  error(0, errno, "strdup");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -659,7 +661,7 @@
 	custom_argv = realloc(custom_argv, sizeof(char *)
 			      * ((unsigned int) custom_argc + 1));
 	if(custom_argv == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  exitstatus = EX_OSERR;
 	  free(org_line);
 	  goto fallback;
@@ -672,7 +674,7 @@
       ret = fclose(conffp);
     } while(ret == EOF and errno == EINTR);
     if(ret == EOF){
-      perror("fclose");
+      error(0, errno, "fclose");
       exitstatus = EX_IOERR;
       goto fallback;
     }
@@ -681,7 +683,7 @@
     /* Check for harmful errors and go to fallback. Other errors might
        not affect opening plugins */
     if(errno == EMFILE or errno == ENFILE or errno == ENOMEM){
-      perror("fopen");
+      error(0, errno, "fopen");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -698,7 +700,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       exitstatus = EX_OSERR;
       goto fallback;
     case EINVAL:
@@ -718,7 +720,7 @@
   case ENOMEM:
   default:
     errno = ret;
-    perror("argp_parse");
+    error(0, errno, "argp_parse");
     exitstatus = EX_OSERR;
     goto fallback;
   case EINVAL:
@@ -743,11 +745,11 @@
   /* Strip permissions down to nobody */
   setgid(gid);
   if(ret == -1){
-    perror("setgid");
+    error(0, errno, "setgid");
   }
   ret = setuid(uid);
   if(ret == -1){
-    perror("setuid");
+    error(0, errno, "setuid");
   }
   
   /* Open plugin directory with close_on_exec flag */
@@ -771,7 +773,7 @@
 		    );
     }
     if(dir_fd == -1){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
       goto fallback;
     }
@@ -780,7 +782,7 @@
   /* Set the FD_CLOEXEC flag on the directory */
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -789,7 +791,7 @@
     
     dir = fdopendir(dir_fd);
     if(dir == NULL){
-      perror("Could not open plugin dir");
+      error(0, errno, "Could not open plugin dir");
       TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
@@ -807,7 +809,7 @@
     /* All directory entries have been processed */
     if(dirst == NULL){
       if(errno == EBADF){
-	perror("readdir");
+	error(0, errno, "readdir");
 	exitstatus = EX_IOERR;
 	goto fallback;
       }
@@ -870,13 +872,13 @@
 					     dirst->d_name));
     }
     if(ret < 0){
-      perror("asprintf");
+      error(0, errno, "asprintf");
       continue;
     }
     
     ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
     if(ret == -1){
-      perror("stat");
+      error(0, errno, "stat");
       free(filename);
       continue;
     }
@@ -894,7 +896,7 @@
     
     plugin *p = getplugin(dirst->d_name);
     if(p == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       free(filename);
       continue;
     }
@@ -912,13 +914,13 @@
       if(g != NULL){
 	for(char **a = g->argv + 1; *a != NULL; a++){
 	  if(not add_argument(p, *a)){
-	    perror("add_argument");
+	    error(0, errno, "add_argument");
 	  }
 	}
 	/* Add global environment variables */
 	for(char **e = g->environ; *e != NULL; e++){
 	  if(not add_environment(p, *e, false)){
-	    perror("add_environment");
+	    error(0, errno, "add_environment");
 	  }
 	}
       }
@@ -929,7 +931,7 @@
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
-	  perror("add_environment");
+	  error(0, errno, "add_environment");
 	}
       }
     }
@@ -937,20 +939,20 @@
     int pipefd[2];
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
     if(ret == -1){
-      perror("pipe");
+      error(0, errno, "pipe");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
-      perror("set_cloexec_flag");
+      error(0, errno, "set_cloexec_flag");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -959,7 +961,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -969,7 +971,7 @@
       pid = fork();
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
-      perror("fork");
+      error(0, errno, "fork");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -977,18 +979,18 @@
       /* this is the child process */
       ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
       if(ret < 0){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	_exit(EX_OSERR);
       }
       ret = sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
       if(ret < 0){
-	perror("sigprocmask");
+	error(0, errno, "sigprocmask");
 	_exit(EX_OSERR);
       }
       
       ret = dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
@@ -999,12 +1001,12 @@
       }
       if(p->environ[0] == NULL){
 	if(execv(filename, p->argv) < 0){
-	  perror("execv");
+	  error(0, errno, "execv for %s", filename);
 	  _exit(EX_OSERR);
 	}
       } else {
 	if(execve(filename, p->argv, p->environ) < 0){
-	  perror("execve");
+	  error(0, errno, "execve for %s", filename);
 	  _exit(EX_OSERR);
 	}
       }
@@ -1016,12 +1018,12 @@
     free(filename);
     plugin *new_plugin = getplugin(dirst->d_name);
     if(new_plugin == NULL){
-      perror("getplugin");
+      error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
 		  (sigprocmask(SIG_UNBLOCK, &sigchld_action.sa_mask,
 			       NULL)));
       if(ret < 0){
-        perror("sigprocmask");
+        error(0, errno, "sigprocmask");
       }
       exitstatus = EX_OSERR;
       goto fallback;
@@ -1036,7 +1038,7 @@
 					      &sigchld_action.sa_mask,
 					      NULL));
     if(ret < 0){
-      perror("sigprocmask");
+      error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1069,7 +1071,7 @@
     fd_set rfds = rfds_all;
     int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
     if(select_ret == -1 and errno != EINTR){
-      perror("select");
+      error(0, errno, "select");
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1111,7 +1113,7 @@
 					 &sigchld_action.sa_mask,
 					 NULL));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1125,7 +1127,7 @@
 		      (sigprocmask(SIG_UNBLOCK,
 				   &sigchld_action.sa_mask, NULL)));
 	  if(ret < 0){
-	    perror("sigprocmask");
+	    error(0, errno, "sigprocmask");
 	    exitstatus = EX_OSERR;
 	    goto fallback;
 	  }
@@ -1142,7 +1144,7 @@
 	bool bret = print_out_password(proc->buffer,
 				       proc->buffer_length);
 	if(not bret){
-	  perror("print_out_password");
+	  error(0, errno, "print_out_password");
 	  exitstatus = EX_IOERR;
 	}
 	goto fallback;
@@ -1161,7 +1163,7 @@
 	proc->buffer = realloc(proc->buffer, proc->buffer_size
 			       + (size_t) BUFFER_SIZE);
 	if(proc->buffer == NULL){
-	  perror("malloc");
+	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
@@ -1204,7 +1206,7 @@
     }
     bret = print_out_password(passwordbuffer, len);
     if(not bret){
-      perror("print_out_password");
+      error(0, errno, "print_out_password");
       exitstatus = EX_IOERR;
     }
   }
@@ -1212,7 +1214,7 @@
   /* Restore old signal handler */
   ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     exitstatus = EX_OSERR;
   }
   
@@ -1234,7 +1236,7 @@
       ret = kill(p->pid, SIGTERM);
       if(ret == -1 and errno != ESRCH){
 	/* Set-uid proccesses might not get closed */
-	perror("kill");
+	error(0, errno, "kill");
       }
     }
   }
@@ -1244,7 +1246,7 @@
     ret = wait(NULL);
   } while(ret >= 0);
   if(errno != ECHILD){
-    perror("wait");
+    error(0, errno, "wait");
   }
   
   free_plugin_list();

=== modified file 'plugins.d/askpass-fifo.c'
--- plugins.d/askpass-fifo.c	2009-10-18 16:09:05 +0000
+++ plugins.d/askpass-fifo.c	2010-09-07 16:48:58 +0000
@@ -31,7 +31,7 @@
 				   ENOENT, EEXIST, EFAULT, EMFILE,
 				   ENFILE, ENOMEM, EBADF, EINVAL, EIO,
 				   EISDIR, EFBIG */
-#include <stdio.h>		/* perror() */
+#include <error.h>		/* error() */
 #include <stdlib.h>		/* EXIT_FAILURE, NULL, size_t, free(),
 				   realloc(), EXIT_SUCCESS */
 #include <fcntl.h>		/* open(), O_RDONLY */
@@ -51,7 +51,7 @@
   ret = mkfifo(passfifo, S_IRUSR | S_IWUSR);
   if(ret == -1){
     int e = errno;
-    perror("mkfifo");
+    error(0, errno, "mkfifo");
     switch(e){
     case EACCES:
     case ENOTDIR:
@@ -73,7 +73,7 @@
   int fifo_fd = open(passfifo, O_RDONLY);
   if(fifo_fd == -1){
     int e = errno;
-    perror("open");
+    error(0, errno, "open");
     switch(e){
     case EACCES:
     case ENOENT:
@@ -101,7 +101,7 @@
       if(buf_len + blocksize > buf_allocated){
 	char *tmp = realloc(buf, buf_allocated + blocksize);
 	if(tmp == NULL){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  free(buf);
 	  return EX_OSERR;
 	}
@@ -113,7 +113,7 @@
 	int e = errno;
 	free(buf);
 	errno = e;
-	perror("read");
+	error(0, errno, "read");
 	switch(e){
 	case EBADF:
 	case EFAULT:
@@ -141,7 +141,7 @@
       int e = errno;
       free(buf);
       errno = e;
-      perror("write");
+      error(0, errno, "write");
       switch(e){
       case EBADF:
       case EFAULT:
@@ -161,7 +161,7 @@
   ret = close(STDOUT_FILENO);
   if(ret == -1){
     int e = errno;
-    perror("close");
+    error(0, errno, "close");
     switch(e){
     case EBADF:
       return EX_OSFILE;

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2010-09-09 18:16:14 +0000
+++ plugins.d/mandos-client.c	2010-09-12 18:12:11 +0000
@@ -63,7 +63,7 @@
 				   strtoimax() */
 #include <assert.h>		/* assert() */
 #include <errno.h>		/* perror(), errno */
-#include <time.h>		/* nanosleep(), time() */
+#include <time.h>		/* nanosleep(), time(), sleep() */
 #include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
 				   SIOCSIFFLAGS, if_indextoname(),
 				   if_nametoindex(), IF_NAMESIZE */
@@ -690,9 +690,11 @@
     ret = connect(tcp_sd, &to.in, sizeof(to)); /* IPv4 */
   }
   if(ret < 0){
-    int e = errno;
-    perror("connect");
-    errno = e;
+    if ((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
+      int e = errno;
+      perror("connect");
+      errno = e;
+    }
     goto mandos_end;
   }
   
@@ -1618,31 +1620,19 @@
     if(quit_now){
       goto end;
     }
-    
-    ret = start_mandos_communication(address, port, if_index, af);
-    if(ret < 0){
-      switch(errno){
-      case ENETUNREACH:
-      case EHOSTDOWN:
-      case EHOSTUNREACH:
-	exitcode = EX_NOHOST;
-	break;
-      case EINVAL:
-	exitcode = EX_USAGE;
-	break;
-      case EIO:
-	exitcode = EX_IOERR;
-	break;
-      case EPROTO:
-	exitcode = EX_PROTOCOL;
-	break;
-      default:
-	exitcode = EX_OSERR;
+
+    while(not quit_now){
+      ret = start_mandos_communication(address, port, if_index, af);
+      if(quit_now or ret == 0){
 	break;
       }
-    } else {
+      sleep(15);
+    };
+
+    if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
+
     goto end;
   }
   

=== modified file 'plugins.d/password-prompt.c'
--- plugins.d/password-prompt.c	2010-03-27 18:39:02 +0000
+++ plugins.d/password-prompt.c	2010-09-07 16:48:58 +0000
@@ -39,11 +39,12 @@
 #include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
 				   getenv() */
 #include <stdio.h>		/* fprintf(), stderr, getline(),
-				   stdin, feof(), perror(), fputc()
+				   stdin, feof(), fputc()
 				*/
 #include <errno.h>		/* errno, EBADF, ENOTTY, EINVAL,
 				   EFAULT, EFBIG, EIO, ENOSPC, EINTR
 				*/
+#include <error.h>		/* error() */
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h> 		/* strlen, rindex */
@@ -70,7 +71,8 @@
 }
 
 int main(int argc, char **argv){
-  ssize_t ret;
+  ssize_t sret;
+  int ret;
   size_t n;
   struct termios t_new, t_old;
   char *buffer = NULL;
@@ -139,7 +141,7 @@
     case ENOMEM:
     default:
       errno = ret;
-      perror("argp_parse");
+      error(0, errno, "argp_parse");
       return EX_OSERR;
     case EINVAL:
       return EX_USAGE;
@@ -155,7 +157,7 @@
   
   if(tcgetattr(STDIN_FILENO, &t_old) != 0){
     int e = errno;
-    perror("tcgetattr");
+    error(0, errno, "tcgetattr");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -168,17 +170,17 @@
   sigemptyset(&new_action.sa_mask);
   ret = sigaddset(&new_action.sa_mask, SIGINT);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGHUP);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   ret = sigaddset(&new_action.sa_mask, SIGTERM);
   if(ret == -1){
-    perror("sigaddset");
+    error(0, errno, "sigaddset");
     return EX_OSERR;
   }
   /* Need to check if the handler is SIG_IGN before handling:
@@ -187,37 +189,37 @@
   */
   ret = sigaction(SIGINT, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGINT, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGHUP, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGHUP, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
   ret = sigaction(SIGTERM, NULL, &old_action);
   if(ret == -1){
-    perror("sigaction");
+    error(0, errno, "sigaction");
     return EX_OSERR;
   }
   if(old_action.sa_handler != SIG_IGN){
     ret = sigaction(SIGTERM, &new_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       return EX_OSERR;
     }
   }
@@ -231,7 +233,7 @@
   t_new.c_lflag &= ~(tcflag_t)ECHO;
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_new) != 0){
     int e = errno;
-    perror("tcsetattr-echo");
+    error(0, errno, "tcsetattr-echo");
     switch(e){
     case EBADF:
     case ENOTTY:
@@ -286,11 +288,11 @@
 	}
       }
     }
-    ret = getline(&buffer, &n, stdin);
-    if(ret > 0){
+    sret = getline(&buffer, &n, stdin);
+    if(sret > 0){
       status = EXIT_SUCCESS;
       /* Make n = data size instead of allocated buffer size */
-      n = (size_t)ret;
+      n = (size_t)sret;
       /* Strip final newline */
       if(n > 0 and buffer[n-1] == '\n'){
 	buffer[n-1] = '\0';	/* not strictly necessary */
@@ -298,10 +300,10 @@
       }
       size_t written = 0;
       while(written < n){
-	ret = write(STDOUT_FILENO, buffer + written, n - written);
-	if(ret < 0){
+	sret = write(STDOUT_FILENO, buffer + written, n - written);
+	if(sret < 0){
 	  int e = errno;
-	  perror("write");
+	  error(0, errno, "write");
 	  switch(e){
 	  case EBADF:
 	  case EFAULT:
@@ -318,12 +320,12 @@
 	  }
 	  break;
 	}
-	written += (size_t)ret;
+	written += (size_t)sret;
       }
-      ret = close(STDOUT_FILENO);
-      if(ret == -1){
+      sret = close(STDOUT_FILENO);
+      if(sret == -1){
 	int e = errno;
-	perror("close");
+	error(0, errno, "close");
 	switch(e){
 	case EBADF:
 	  status = EX_OSFILE;
@@ -336,10 +338,10 @@
       }
       break;
     }
-    if(ret < 0){
+    if(sret < 0){
       int e = errno;
       if(errno != EINTR and not feof(stdin)){
-	perror("getline");
+	error(0, errno, "getline");
 	switch(e){
 	case EBADF:
 	  status = EX_UNAVAILABLE;
@@ -352,7 +354,7 @@
 	break;
       }
     }
-    /* if(ret == 0), then the only sensible thing to do is to retry to
+    /* if(sret == 0), then the only sensible thing to do is to retry to
        read from stdin */
     fputc('\n', stderr);
     if(debug and not quit_now){
@@ -368,7 +370,7 @@
     fprintf(stderr, "Restoring terminal attributes\n");
   }
   if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
-    perror("tcsetattr+echo");
+    error(0, errno, "tcsetattr+echo");
   }
   
   if(quit_now){
@@ -376,7 +378,7 @@
     old_action.sa_handler = SIG_DFL;
     ret = sigaction(signal_received, &old_action, NULL);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     raise(signal_received);
   }

=== added file 'plugins.d/plymouth.c'
--- plugins.d/plymouth.c	1970-01-01 00:00:00 +0000
+++ plugins.d/plymouth.c	2010-09-24 18:33:39 +0000
@@ -0,0 +1,407 @@
+#define _GNU_SOURCE		/* asprintf(), TEMP_FAILURE_RETRY() */
+#include <signal.h>		/* sig_atomic_t, struct sigaction,
+				   sigemptyset(), sigaddset(), SIGINT,
+				   SIGHUP, SIGTERM, sigaction(),
+				   kill(), SIG_IGN */
+#include <stdbool.h>		/* bool, false, true */
+#include <fcntl.h>		/* open(), O_RDONLY */
+#include <iso646.h>		/* and, or, not*/
+#include <sys/types.h>		/* size_t, ssize_t, pid_t, struct
+				   dirent, waitpid() */
+#include <sys/wait.h>		/* waitpid() */
+#include <stddef.h>		/* NULL */
+#include <string.h>		/* strchr(), memcmp() */
+#include <stdio.h>		/* asprintf(), perror(), fopen(),
+				   fscanf() */
+#include <unistd.h>		/* close(), readlink(), read(),
+				   fork(), setsid(), chdir(), dup2(),
+				   STDERR_FILENO, execv(), access() */
+#include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
+#include <dirent.h>		/* scandir(), alphasort() */
+#include <inttypes.h>		/* intmax_t, strtoumax(), SCNuMAX */
+#include <sys/stat.h>		/* struct stat, lstat() */
+#include <sysexits.h>		/* EX_OSERR, EX_UNAVAILABLE */
+#include <error.h>		/* error() */
+#include <errno.h>		/* TEMP_FAILURE_RETRY */
+#include <argz.h>		/* argz_count(), argz_extract() */
+
+sig_atomic_t interrupted_by_signal = 0;
+const char plymouth_pid[] = "/dev/.initramfs/plymouth.pid";
+const char plymouth_path[] = "/bin/plymouth";
+const char plymouthd_path[] = "/sbin/plymouthd";
+const char *plymouthd_default_argv[] = {"/sbin/plymouthd",
+					"--mode=boot",
+					"--attach-to-session",
+					"--pid-file="
+					"/dev/.initramfs/"
+					"plymouth.pid",
+					NULL };
+
+static void termination_handler(__attribute__((unused))int signum){
+  if(interrupted_by_signal){
+    return;
+  }
+  interrupted_by_signal = 1;
+}
+
+/* Create prompt string */
+char *makeprompt(void){
+  int ret = 0;
+  char *prompt;
+  const char *const cryptsource = getenv("cryptsource");
+  const char *const crypttarget = getenv("crypttarget");
+  const char prompt_start[] = "Enter passphrase to unlock the disk";
+  
+  if(cryptsource == NULL){
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s: ", prompt_start);
+    } else {
+      ret = asprintf(&prompt, "%s (%s): ", prompt_start,
+		     crypttarget);
+    }
+  } else {
+    if(crypttarget == NULL){
+      ret = asprintf(&prompt, "%s %s: ", prompt_start, cryptsource);
+    } else {
+      ret = asprintf(&prompt, "%s %s (%s): ", prompt_start,
+		     cryptsource, crypttarget);
+    }
+  }
+  if(ret == -1){
+    return NULL;
+  }
+  return prompt;
+}
+
+void kill_and_wait(pid_t pid){
+  TEMP_FAILURE_RETRY(kill(pid, SIGTERM));
+  TEMP_FAILURE_RETRY(waitpid(pid, NULL, 0));
+}
+
+bool become_a_daemon(void){
+  int ret = setuid(geteuid());
+  if(ret == -1){
+    error(0, errno, "setuid");
+  }
+    
+  setsid();
+  ret = chdir("/");
+  if(ret == -1){
+    error(0, errno, "chdir");
+    return false;
+  }
+  ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
+  if(ret == -1){
+    error(0, errno, "dup2");
+    return false;
+  }
+  return true;
+}
+
+bool exec_and_wait(pid_t *pid_return, const char *path,
+		   const char **argv, bool interruptable,
+		   bool daemonize){
+  int status;
+  int ret;
+  pid_t pid;
+  pid = fork();
+  if(pid == -1){
+    error(0, errno, "fork");
+    return false;
+  }
+  if(pid == 0){
+    /* Child */
+    if(daemonize){
+      if(not become_a_daemon()){
+	_exit(EX_OSERR);
+      }
+    }
+
+    char **new_argv = NULL;
+    char *tmp;
+    int i = 0;
+    for (; argv[i]!=(char *)NULL; i++){
+      tmp = realloc(new_argv, sizeof(const char *) * ((size_t)i + 1));
+      if (tmp == NULL){
+	error(0, errno, "realloc");
+	free(new_argv);
+	_exit(EX_OSERR);
+      }
+      new_argv = (char **)tmp;
+      new_argv[i] = strdup(argv[i]);
+    }
+    new_argv[i] = (char *) NULL;
+    
+    execv(path, (char *const *)new_argv);
+    error(0, errno, "execv");
+    _exit(EXIT_FAILURE);
+  }
+  if(pid_return != NULL){
+    *pid_return = pid;
+  }
+  do {
+    ret = waitpid(pid, &status, 0);
+  } while(ret == -1 and errno == EINTR
+	  and ((not interrupted_by_signal)
+	       or (not interruptable)));
+  if(interrupted_by_signal and interruptable){
+    return false;
+  }
+  if(ret == -1){
+    error(0, errno, "waitpid");
+    return false;
+  }
+  if(WIFEXITED(status) and (WEXITSTATUS(status) == 0)){
+    return true;
+  }
+  return false;
+}
+
+int is_plymouth(const struct dirent *proc_entry){
+  int ret;
+  {
+    uintmax_t maxvalue;
+    char *tmp;
+    errno = 0;
+    maxvalue = strtoumax(proc_entry->d_name, &tmp, 10);
+
+    if(errno != 0 or *tmp != '\0'
+       or maxvalue != (uintmax_t)((pid_t)maxvalue)){
+      return 0;
+    }
+  }
+  char exe_target[sizeof(plymouth_path)];
+  char *exe_link;
+  ret = asprintf(&exe_link, "/proc/%s/exe", proc_entry->d_name);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return 0;
+  }
+  
+  struct stat exe_stat;
+  ret = lstat(exe_link, &exe_stat);
+  if(ret == -1){
+    free(exe_link);
+    if(errno != ENOENT){
+      error(0, errno, "lstat");
+    }
+    return 0;
+  }
+  
+  if(not S_ISLNK(exe_stat.st_mode)
+     or exe_stat.st_uid != 0
+     or exe_stat.st_gid != 0){
+    free(exe_link);
+    return 0;
+  }
+  
+  ssize_t sret = readlink(exe_link, exe_target, sizeof(exe_target));
+  free(exe_link);
+  if((sret != (ssize_t)sizeof(plymouth_path)-1) or
+      (memcmp(plymouth_path, exe_target,
+	      sizeof(plymouth_path)-1) != 0)){
+    return 0;
+  }
+  return 1;
+}
+
+pid_t get_pid(void){
+  int ret;
+  FILE *pidfile = fopen(plymouth_pid, "r");
+  uintmax_t maxvalue = 0;
+  if(pidfile != NULL){
+    ret = fscanf(pidfile, "%" SCNuMAX, &maxvalue);
+    if(ret != 1){
+      maxvalue = 0;
+    }
+    fclose(pidfile);
+  }
+  if(maxvalue == 0){
+    struct dirent **direntries;
+    ret = scandir("/proc", &direntries, is_plymouth, alphasort);
+    sscanf(direntries[0]->d_name, "%" SCNuMAX, &maxvalue);
+  }
+  pid_t pid;
+  pid = (pid_t)maxvalue;
+  if((uintmax_t)pid == maxvalue){
+    return pid;
+  }
+  
+  return 0;
+}
+
+const char **getargv(pid_t pid){
+  int cl_fd;
+  char *cmdline_filename;
+  ssize_t sret;
+  int ret;
+  
+  ret = asprintf(&cmdline_filename, "/proc/%" PRIuMAX "/cmdline",
+		 (uintmax_t)pid);
+  if(ret == -1){
+    error(0, errno, "asprintf");
+    return NULL;
+  }
+  
+  /* Open /proc/<pid>/cmdline  */
+  cl_fd = open(cmdline_filename, O_RDONLY);
+  free(cmdline_filename);
+  if(cl_fd == -1){
+    error(0, errno, "open");
+    return NULL;
+  }
+  
+  size_t cmdline_allocated = 0;
+  size_t cmdline_len = 0;
+  char *cmdline = NULL;
+  char *tmp;
+  const size_t blocksize = 1024;
+  do {
+    /* Allocate more space? */
+    if(cmdline_len + blocksize > cmdline_allocated){
+      tmp = realloc(cmdline, cmdline_allocated + blocksize);
+      if(tmp == NULL){
+	error(0, errno, "realloc");
+	free(cmdline);
+	close(cl_fd);
+	return NULL;
+      }
+      cmdline = tmp;
+      cmdline_allocated += blocksize;
+    }
+    
+    /* Read data */
+    sret = read(cl_fd, cmdline + cmdline_len,
+		cmdline_allocated - cmdline_len);
+    if(sret == -1){
+      error(0, errno, "read");
+      free(cmdline);
+      close(cl_fd);
+      return NULL;
+    }
+    cmdline_len += (size_t)sret;
+  } while(sret != 0);
+  ret = close(cl_fd);
+  if(ret == -1){
+    error(0, errno, "close");
+    free(cmdline);
+    return NULL;
+  }
+  
+  /* we got cmdline and cmdline_len, ignore rest... */
+  char **argv = malloc((argz_count(cmdline, cmdline_len) + 1)
+		       * sizeof(char *)); /* Get number of args */
+  if(argv == NULL){
+    error(0, errno, "argv = malloc()");
+    free(cmdline);
+    return NULL;
+  }
+  argz_extract(cmdline, cmdline_len, argv); /* Create argv */
+  return (const char **)argv;
+}
+
+int main(__attribute__((unused))int argc,
+	 __attribute__((unused))char **argv){
+  char *prompt;
+  char *prompt_arg;
+  pid_t plymouth_command_pid;
+  int ret;
+  bool bret;
+
+  /* test -x /bin/plymouth */
+  ret = access(plymouth_path, X_OK);
+  if(ret == -1){
+    /* Plymouth is probably not installed.  Don't print an error
+       message, just exit. */
+    exit(EX_UNAVAILABLE);
+  }
+  
+  { /* Add signal handlers */
+    struct sigaction old_action,
+      new_action = { .sa_handler = termination_handler,
+		     .sa_flags = 0 };
+    sigemptyset(&new_action.sa_mask);
+    for(int *sig = (int[]){ SIGINT, SIGHUP, SIGTERM, 0 };
+	*sig != 0; sig++){
+      ret = sigaddset(&new_action.sa_mask, *sig);
+      if(ret == -1){
+	error(EX_OSERR, errno, "sigaddset");
+      }
+      ret = sigaction(*sig, NULL, &old_action);
+      if(ret == -1){
+	error(EX_OSERR, errno, "sigaction");
+      }
+      if(old_action.sa_handler != SIG_IGN){
+	ret = sigaction(*sig, &new_action, NULL);
+	if(ret == -1){
+	  error(EX_OSERR, errno, "sigaction");
+	}
+      }
+    }
+  }
+  
+  /* plymouth --ping */
+  bret = exec_and_wait(&plymouth_command_pid, plymouth_path,
+		       (const char *[])
+		       { plymouth_path, "--ping", NULL },
+		       true, false);
+  if(not bret){
+    if(interrupted_by_signal){
+      kill_and_wait(plymouth_command_pid);
+      exit(EXIT_FAILURE);
+    }
+    /* Plymouth is probably not running.  Don't print an error
+       message, just exit. */
+    exit(EX_UNAVAILABLE);
+  }
+  
+  prompt = makeprompt();
+  ret = asprintf(&prompt_arg, "--prompt=%s", prompt);
+  free(prompt);
+  if(ret == -1){
+    error(EX_OSERR, errno, "asprintf");
+  }
+  
+  /* plymouth ask-for-password --prompt="$prompt" */
+  bret = exec_and_wait(&plymouth_command_pid,
+		       plymouth_path, (const char *[])
+		       { plymouth_path, "ask-for-password",
+			   prompt_arg, NULL },
+		       true, false);
+  free(prompt_arg);
+  if(bret){
+    exit(EXIT_SUCCESS);
+  }
+  if(not interrupted_by_signal){
+    /* exec_and_wait failed for some other reason */
+    exit(EXIT_FAILURE);
+  }
+  kill_and_wait(plymouth_command_pid);
+  
+  const char **plymouthd_argv;
+  pid_t pid = get_pid();
+  if(pid == 0){
+    error(0, 0, "plymouthd pid not found");
+    plymouthd_argv = plymouthd_default_argv;
+  } else {
+    plymouthd_argv = getargv(pid);
+  }
+  
+  bret = exec_and_wait(NULL, plymouth_path, (const char *[])
+		       { plymouth_path, "quit", NULL },
+		       false, false);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  bret = exec_and_wait(NULL, plymouthd_path, plymouthd_argv,
+		       false, true);
+  if(not bret){
+    exit(EXIT_FAILURE);
+  }
+  exec_and_wait(NULL, plymouth_path, (const char *[])
+		{ plymouth_path, "show-splash", NULL },
+		false, false);
+  exit(EXIT_FAILURE);
+}

=== added file 'plugins.d/plymouth.xml'
--- plugins.d/plymouth.xml	1970-01-01 00:00:00 +0000
+++ plugins.d/plymouth.xml	2010-09-23 20:14:08 +0000
@@ -0,0 +1,279 @@
+<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY COMMANDNAME "plymouth">
+<!ENTITY TIMESTAMP "2010-09-23">
+<!ENTITY % common SYSTEM "../common.ent">
+%common;
+]>
+
+<refentry xmlns:xi="http://www.w3.org/2001/XInclude">
+  <refentryinfo>
+    <title>Mandos Manual</title>
+    <!-- NWalsh’s docbook scripts use this to generate the footer: -->
+    <productname>Mandos</productname>
+    <productnumber>&version;</productnumber>
+    <date>&TIMESTAMP;</date>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <year>2009</year>
+      <holder>Teddy Hogeborn</holder>
+      <holder>Björn Påhlsson</holder>
+    </copyright>
+    <xi:include href="../legalnotice.xml"/>
+  </refentryinfo>
+  
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8mandos</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>Mandos plugin to use plymouth to get a
+    password.</refpurpose>
+  </refnamediv>
+  
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+    </cmdsynopsis>
+  </refsynopsisdiv>
+  
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      This program prompts for a password using <citerefentry>
+      <refentrytitle>plymouth</refentrytitle><manvolnum>8</manvolnum>
+      </citerefentry> and outputs any given password to standard
+      output.  If no <citerefentry><refentrytitle
+      >plymouth</refentrytitle><manvolnum>8</manvolnum></citerefentry>
+      process can be found, this program will immediately exit with an
+      exit code indicating failure.
+    </para>
+    <para>
+      This program is not very useful on its own.  This program is
+      really meant to run as a plugin in the <application
+      >Mandos</application> client-side system, where it is used as a
+      fallback and alternative to retrieving passwords from a
+      <application >Mandos</application> server.
+    </para>
+    <para>
+      If this program is killed (presumably by
+      <citerefentry><refentrytitle>plugin-runner</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry> because some other
+      plugin provided the password), it cannot tell <citerefentry>
+      <refentrytitle>plymouth</refentrytitle><manvolnum>8</manvolnum>
+      </citerefentry> to abort requesting a password, because
+      <citerefentry><refentrytitle>plymouth</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry> does not support this.
+      Therefore, this program will then <emphasis>kill</emphasis> the
+      running <citerefentry><refentrytitle>plymouth</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry> process and start a
+      <emphasis>new</emphasis> one using the same command line
+      arguments as the old one was using.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="options">
+    <title>OPTIONS</title>
+    <para>
+      This program takes no options.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="exit_status">
+    <title>EXIT STATUS</title>
+    <para>
+      If exit status is 0, the output from the program is the password
+      as it was read.  Otherwise, if exit status is other than 0, the
+      program was interrupted or encountered an error, and any output
+      so far could be corrupt and/or truncated, and should therefore
+      be ignored.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="environment">
+    <title>ENVIRONMENT</title>
+    <variablelist>
+      <varlistentry>
+	<term><envar>cryptsource</envar></term>
+	<term><envar>crypttarget</envar></term>
+	<listitem>
+	  <para>
+	    If set, these environment variables will be assumed to
+	    contain the source device name and the target device
+	    mapper name, respectively, and will be shown as part of
+	    the prompt.
+	</para>
+	<para>
+	  These variables will normally be inherited from
+	  <citerefentry><refentrytitle>plugin-runner</refentrytitle>
+	  <manvolnum>8mandos</manvolnum></citerefentry>, which will
+	  normally have inherited them from
+	  <filename>/scripts/local-top/cryptroot</filename> in the
+	  initial <acronym>RAM</acronym> disk environment, which will
+	  have set them from parsing kernel arguments and
+	  <filename>/conf/conf.d/cryptroot</filename> (also in the
+	  initial RAM disk environment), which in turn will have been
+	  created when the initial RAM disk image was created by
+	  <filename
+	  >/usr/share/initramfs-tools/hooks/cryptroot</filename>, by
+	  extracting the information of the root file system from
+	  <filename >/etc/crypttab</filename>.
+	</para>
+	<para>
+	  This behavior is meant to exactly mirror the behavior of
+	  <command>askpass</command>, the default password prompter.
+	</para>
+	</listitem>
+      </varlistentry>
+    </variablelist>
+  </refsect1>
+  
+  <refsect1 id="files">
+    <title>FILES</title>
+    <variablelist>
+      <varlistentry>
+	<term><filename>/bin/plymouth</filename></term>
+	<listitem>
+	  <para>
+	    This is the command run to retrieve a password from
+	    <citerefentry><refentrytitle>plymouth</refentrytitle>
+	    <manvolnum>8</manvolnum></citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      <varlistentry>
+	<term><filename>/proc</filename></term>
+	<listitem>
+	  <para>
+	    To find the running <citerefentry><refentrytitle
+	    >plymouth</refentrytitle><manvolnum>8</manvolnum>
+	    </citerefentry>, this directory will be searched for
+	    numeric entries which will be assumed to be directories.
+	    In all those directories, the <filename>exe</filename> and
+	    <filename>cmdline</filename> entries will be used to
+	    determine the name of the running binary, effective user
+	    and group <abbrev>ID</abbrev>, and the command line
+	    arguments.  See <citerefentry><refentrytitle
+	    >proc</refentrytitle><manvolnum>5</manvolnum>
+	    </citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+      <varlistentry>
+	<term><filename>/sbin/plymouthd</filename></term>
+	<listitem>
+	  <para>
+	    This is the name of the binary which will be searched for
+	    in the process list.  See <citerefentry><refentrytitle
+	    >plymouth</refentrytitle><manvolnum>8</manvolnum>
+	    </citerefentry>.
+	  </para>
+	</listitem>
+      </varlistentry>
+    </variablelist>
+  </refsect1>
+  
+  <refsect1 id="bugs">
+    <title>BUGS</title>
+    <para>
+      Killing the <citerefentry><refentrytitle
+      >plymouth</refentrytitle><manvolnum>8</manvolnum></citerefentry>
+      daemon and starting a new one is ugly, but necessary as long as
+      it does not support aborting a password request.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="example">
+    <title>EXAMPLE</title>
+    <para>
+      Note that normally, this program will not be invoked directly,
+      but instead started by the Mandos <citerefentry><refentrytitle
+      >plugin-runner</refentrytitle><manvolnum>8mandos</manvolnum>
+      </citerefentry>.
+    </para>
+    <informalexample>
+      <para>
+	This program takes no options.
+      </para>
+      <para>
+	<userinput>&COMMANDNAME;</userinput>
+      </para>
+    </informalexample>
+  </refsect1>
+  
+  <refsect1 id="security">
+    <title>SECURITY</title>
+    <para>
+      If this program is killed by a signal, it will kill the process
+      <abbrev>ID</abbrev> which at the start of this program was
+      determined to run <citerefentry><refentrytitle
+      >plymouth</refentrytitle><manvolnum>8</manvolnum></citerefentry>
+      as root (see also <xref linkend="files"/>).  There is a very
+      slight risk that, in the time between those events, that process
+      <abbrev>ID</abbrev> was freed and then taken up by another
+      process; the wrong process would then be killed.  Now, this
+      program can only be killed by the user who started it; see
+      <citerefentry><refentrytitle>plugin-runner</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry>.  This program
+      should therefore be started by a completely separate
+      non-privileged user, and no other programs should be allowed to
+      run as that special user.  This means that it is not recommended
+      to use the user "nobody" to start this program, as other
+      possibly less trusted programs could be running as "nobody", and
+      they would then be able to kill this program, triggering the
+      killing of the process <abbrev>ID</abbrev> which may or may not
+      be <citerefentry><refentrytitle>plymouth</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>.
+    </para>
+    <para>
+      The only other thing that could be considered worthy of note is
+      this:  This program is meant to be run by <citerefentry>
+      <refentrytitle>plugin-runner</refentrytitle><manvolnum
+      >8mandos</manvolnum></citerefentry>, and will, when run
+      standalone, outside, in a normal environment, immediately output
+      on its standard output any presumably secret password it just
+      received.  Therefore, when running this program standalone
+      (which should never normally be done), take care not to type in
+      any real secret password by force of habit, since it would then
+      immediately be shown as output.
+    </para>
+  </refsect1>
+  
+  <refsect1 id="see_also">
+    <title>SEE ALSO</title>
+    <para>
+      <citerefentry><refentrytitle>crypttab</refentrytitle>
+      <manvolnum>5</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>plugin-runner</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>proc</refentrytitle>
+      <manvolnum>5</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>plymouth</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>
+    </para>
+  </refsect1>
+</refentry>
+<!-- Local Variables: -->
+<!-- time-stamp-start: "<!ENTITY TIMESTAMP [\"']" -->
+<!-- time-stamp-end: "[\"']>" -->
+<!-- time-stamp-format: "%:y-%02m-%02d" -->
+<!-- End: -->

=== modified file 'plugins.d/splashy.c'
--- plugins.d/splashy.c	2009-10-18 16:09:05 +0000
+++ plugins.d/splashy.c	2010-09-07 16:48:58 +0000
@@ -29,7 +29,7 @@
 				   SIG_IGN, kill(), SIGKILL */
 #include <stddef.h>		/* NULL */
 #include <stdlib.h>		/* getenv() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf() */
 #include <stdlib.h>		/* EXIT_FAILURE, free(),
 				   EXIT_SUCCESS */
 #include <sys/types.h>		/* pid_t, DIR, struct dirent,
@@ -49,6 +49,7 @@
 				   E2BIG, EFAULT, EIO, ETXTBSY,
 				   EISDIR, ELIBBAD, EPERM, EINTR,
 				   ECHILD */
+#include <error.h>		/* error() */
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS() */
 #include <sysexits.h>		/* EX_OSERR, EX_OSFILE,
@@ -109,7 +110,7 @@
     proc_dir = opendir("/proc");
     if(proc_dir == NULL){
       int e = errno;
-      perror("opendir");
+      error(0, errno, "opendir");
       switch(e){
       case EACCES:
       case ENOTDIR:
@@ -151,7 +152,7 @@
 	char *exe_link;
 	ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  exitstatus = EX_OSERR;
 	  goto failure;
 	}
@@ -165,7 +166,7 @@
 	    continue;
 	  }
 	  int e = errno;
-	  perror("lstat");
+	  error(0, errno, "lstat");
 	  free(exe_link);
 	  switch(e){
 	  case EACCES:
@@ -213,60 +214,60 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
       exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	exitstatus = EX_OSERR;
 	goto failure;
       }
@@ -283,7 +284,7 @@
     goto failure;
   }
   if(splashy_command_pid == -1){
-    perror("fork");
+    error(0, errno, "fork");
     exitstatus = EX_OSERR;
     goto failure;
   }
@@ -293,7 +294,7 @@
       const char splashy_command[] = "/sbin/splashy_update";
       execl(splashy_command, splashy_command, prompt, (char *)NULL);
       int e = errno;
-      perror("execl");
+      error(0, errno, "execl");
       switch(e){
       case EACCES:
       case ENOENT:
@@ -341,7 +342,7 @@
       goto failure;
     }
     if(ret == -1){
-      perror("waitpid");
+      error(0, errno, "waitpid");
       if(errno == ECHILD){
 	splashy_command_pid = 0;
       }
@@ -379,27 +380,27 @@
 	 the real user ID (_mandos) */
       ret = setuid(geteuid());
       if(ret == -1){
-	perror("setuid");
+	error(0, errno, "setuid");
       }
       
       setsid();
       ret = chdir("/");
       if(ret == -1){
-	perror("chdir");
+	error(0, errno, "chdir");
       }
 /*       if(fork() != 0){ */
 /* 	_exit(EXIT_SUCCESS); */
 /*       } */
       ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace stdout */
       if(ret == -1){
-	perror("dup2");
+	error(0, errno, "dup2");
 	_exit(EX_OSERR);
       }
       
       execl("/sbin/splashy", "/sbin/splashy", "boot", (char *)NULL);
       {
 	int e = errno;
-	perror("execl");
+	error(0, errno, "execl");
 	switch(e){
 	case EACCES:
 	case ENOENT:
@@ -425,13 +426,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2010-03-27 18:39:02 +0000
+++ plugins.d/usplash.c	2010-09-15 17:21:04 +0000
@@ -31,23 +31,25 @@
 #include <fcntl.h>		/* open(), O_WRONLY, O_RDONLY */
 #include <iso646.h>		/* and, or, not*/
 #include <errno.h>		/* errno, EINTR */
+#include <error.h>
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
 #include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf()*/
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
 #include <sysexits.h>		/* EX_OSERR, EX_UNAVAILABLE */
+#include <argz.h>		/* argz_count(), argz_extract() */
 
 sig_atomic_t interrupted_by_signal = 0;
 int signal_received;
@@ -153,7 +155,7 @@
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    perror("opendir");
+    error(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -181,7 +183,7 @@
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	perror("asprintf");
+	error(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -193,7 +195,7 @@
 	  free(exe_link);
 	  continue;
 	}
-	perror("lstat");
+	error(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -224,13 +226,13 @@
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  perror("open");
+	  error(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -242,7 +244,7 @@
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    perror("realloc");
+	    error(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -253,7 +255,7 @@
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  perror("read");
+	  error(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -261,14 +263,14 @@
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	perror("close");
+	error(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      perror("closedir");
+      error(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -323,26 +325,26 @@
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
       status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -351,7 +353,7 @@
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -360,7 +362,7 @@
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -369,7 +371,7 @@
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -378,7 +380,7 @@
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
 	status = EX_OSERR;
       }
       goto failure;
@@ -387,7 +389,7 @@
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -399,7 +401,7 @@
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -411,7 +413,7 @@
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -429,7 +431,7 @@
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      perror("open");
+      error(0, errno, "open");
       status = EX_OSERR;
     }
     goto failure;
@@ -448,7 +450,7 @@
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  perror("realloc");
+	  error(0, errno, "realloc");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -460,7 +462,7 @@
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	perror("read");
+	error(0, errno, "read");
 	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
@@ -475,7 +477,7 @@
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -488,7 +490,7 @@
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
       status = EX_OSERR;
     }
     goto failure;
@@ -501,7 +503,7 @@
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
       status = EX_OSERR;
     }
     goto failure;
@@ -515,7 +517,7 @@
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  perror("write");
+	  error(0, errno, "write");
 	  status = EX_OSERR;
 	}
 	goto failure;
@@ -552,7 +554,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -561,31 +563,19 @@
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      perror("close");
-    }
-  }
-  
-  /* Create argc and argv for new usplash*/
-  int cmdline_argc = 0;
-  char **cmdline_argv = malloc(sizeof(char *));
-  {
-    size_t position = 0;
-    while(position < cmdline_len){
-      char **tmp = realloc(cmdline_argv,
-			   (sizeof(char *)
-			    * (size_t)(cmdline_argc + 2)));
-      if(tmp == NULL){
-	perror("realloc");
-	free(cmdline_argv);
-	return status;
-      }
-      cmdline_argv = tmp;
-      cmdline_argv[cmdline_argc] = cmdline + position;
-      cmdline_argc++;
-      position += strlen(cmdline + position) + 1;
-    }
-    cmdline_argv[cmdline_argc] = NULL;
-  }
+      error(0, errno, "close");
+    }
+  }
+  
+  /* Create argv for new usplash*/
+  char **cmdline_argv = malloc((argz_count(cmdline, cmdline_len) + 1)
+			       * sizeof(char *)); /* Count args */
+  if(cmdline_argv == NULL){
+    error(0, errno, "malloc");
+    return status;
+  }
+  argz_extract(cmdline, cmdline_len, cmdline_argv); /* Create argv */
+  
   /* Kill old usplash */
   kill(usplash_pid, SIGTERM);
   sleep(2);
@@ -602,23 +592,27 @@
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      perror("setuid");
+      error(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
+    if(ret == -1){
+      error(0, errno, "chdir");
+      _exit(EX_OSERR);
+    }
 /*     if(fork() != 0){ */
 /*       _exit(EXIT_SUCCESS); */
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      perror("dup2");
+      error(0, errno, "dup2");
       _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      perror("execv");
+      error(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
@@ -629,7 +623,7 @@
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
     }
   }
   
@@ -637,7 +631,7 @@
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -648,13 +642,13 @@
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());

